#!/usr/bin/env python
# encoding: ISO8859-1
# Thomas Nagy, 2005-2016
#
"""
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

import os, sys, inspect

VERSION="1.9.1"
REVISION="584ef1b5b82e22046b7c9d5797d92393"
GIT="1fa2e6f7e1420e3fc64b29bf7abef531220da538"
INSTALL=''
C1='#*'
C2='#&'
C3='#%'
cwd = os.getcwd()
join = os.path.join


WAF='waf'
def b(x):
	return x
if sys.hexversion>0x300000f:
	WAF='waf3'
	def b(x):
		return x.encode()

def err(m):
	print(('\033[91mError: %s\033[0m' % m))
	sys.exit(1)

def unpack_wafdir(dir, src):
	f = open(src,'rb')
	c = 'corrupt archive (%d)'
	while 1:
		line = f.readline()
		if not line: err('run waf-light from a folder containing waflib')
		if line == b('#==>\n'):
			txt = f.readline()
			if not txt: err(c % 1)
			if f.readline() != b('#<==\n'): err(c % 2)
			break
	if not txt: err(c % 3)
	txt = txt[1:-1].replace(b(C1), b('\n')).replace(b(C2), b('\r')).replace(b(C3), b('\x00'))

	import shutil, tarfile
	try: shutil.rmtree(dir)
	except OSError: pass
	try:
		for x in ('Tools', 'extras'):
			os.makedirs(join(dir, 'waflib', x))
	except OSError:
		err("Cannot unpack waf lib into %s\nMove waf in a writable directory" % dir)

	os.chdir(dir)
	tmp = 't.bz2'
	t = open(tmp,'wb')
	try: t.write(txt)
	finally: t.close()

	try:
		t = tarfile.open(tmp)
	except:
		try:
			os.system('bunzip2 t.bz2')
			t = tarfile.open('t')
			tmp = 't'
		except:
			os.chdir(cwd)
			try: shutil.rmtree(dir)
			except OSError: pass
			err("Waf cannot be unpacked, check that bzip2 support is present")

	try:
		for x in t: t.extract(x)
	finally:
		t.close()

	for x in ('Tools', 'extras'):
		os.chmod(join('waflib',x), 493)

	if sys.hexversion<0x300000f:
		sys.path = [join(dir, 'waflib')] + sys.path
		import fixpy2
		fixpy2.fixdir(dir)

	os.remove(tmp)
	os.chdir(cwd)

	try: dir = unicode(dir, 'mbcs')
	except: pass
	try:
		from ctypes import windll
		windll.kernel32.SetFileAttributesW(dir, 2)
	except:
		pass

def test(dir):
	try:
		os.stat(join(dir, 'waflib'))
		return os.path.abspath(dir)
	except OSError:
		pass

def find_lib():
	src = os.path.abspath(inspect.getfile(inspect.getmodule(err)))
	base, name = os.path.split(src)

	#devs use $WAFDIR
	w=test(os.environ.get('WAFDIR', ''))
	if w: return w

	#waf-light
	if name.endswith('waf-light'):
		w = test(base)
		if w: return w
		err('waf-light requires waflib -> export WAFDIR=/folder')

	dirname = '%s-%s-%s' % (WAF, VERSION, REVISION)
	for i in (INSTALL,'/usr','/usr/local','/opt'):
		w = test(i + '/lib/' + dirname)
		if w: return w

	#waf-local
	dir = join(base, (sys.platform != 'win32' and '.' or '') + dirname)
	w = test(dir)
	if w: return w

	#unpack
	unpack_wafdir(dir, src)
	return dir

wafdir = find_lib()
sys.path.insert(0, wafdir)

if __name__ == '__main__':

	from waflib import Scripting
	Scripting.waf_entry_point(cwd, VERSION, wafdir)

#==>
#BZh91AY&SYDpB&#%MC #%E(a\{g2#%#%#%#%#%#%#%#%#%#%#%#%#%#%#%#%#%#%#%#%>y5Zf}vjUl=wfkl5۽/sNz0-A^{ݞ⥩H;u<ؕgεӻK/g[/7w{*^{=}i˽zp=lZ}#%#%#%h#%	f8{{zzC;sjSm@#%(3jva#%#%z4se#%I*GP#%R#*.yt;sw;E=%Kosnޮ]׼%;/}}_'/}y^qr]jnق"iZ+]^V[nt4Q]鍲̈{Ӄ#*zPJMڗY2װh&{k4ݎ>P^Ž֮}ـwvx&w([5{p;w4pe]w}vcgwn.ܷ^Qm<=<xW,k#%74SxRbm@¦|ɵiͶ]]ɵ׶^I۴׮{%Nuk#%yނ>ҋPUH/#&Aw/`׻WАq>n#*z-UUmǼKux4 ]x#%#%#%wy;ٺY78vƎrtSsdlZu+s핡bDOK<N-ۇmۼt{J;i}44}}bw5mn^7:_}Ǹ>]9=rϹLj#&#%/.]T59:+:>m/\{\wg^);5޷oUÕW;op{[<m+j:wXznRݵW<wvq^{ͰCݻӫӠE.ZƠ#&}hfo'gwv8V=Usn3vv۴vΒR,qUM6Wq=oL$͎}>=۵빟"aG=V#&lJF^v}Ƽ]^zo'3|,4#%#%@1&SHS`F@y=PJh	2DSĆ=LCMH#%#%#%#%#%A" &d	4L#*oJ<O&42a#%#%'RDjf~ =M#&#%ځ#&#&#%#%#%#%#%$&@#%FiFmF#%#%#%#%j" #SlT$zSOO#d#%2#%#%:(m5_Ui$\M42#THE>?7"JjL=~i|2\Ӓ\^/OVxP (;]t.8TER0<ԫJMDMM&/F!_ҪmhguuVƭdU2V*,6A(1Ql@@H@D^ZҵM]ѬV5Zֵl2Dfƙ dCM3Q5"Xȋ)FSm&*fI`ɪj6ƃ%l#&d(#Qc@bʛF)6KEhJYiJmDi	(͌kF	hƔ PiR"6Z[ZkjU4f&L 4ڛZ3K@%&J&fMKS-%)%F!lQAL4HQ) ѰBҢ ԥh`DdZIFP#&%4ѓ"5IQbaHEJ`hM&DK&41%DRBjbi1$X&SS2d́"%*I65X`X$ؠ*hDFM)b	10%S*"5	TI0"RlI%%`hƚ$L$E(̈́&mc )`#&4ؤF&d4#%,$T(1fQC#4U	4Lh"6JLiQ)YHD2d&PR1#%l*fVe6!HBfVH(I"JI,Y6lh!RfHK"l`"̚S$)f,hIEJL2bLVMH6R#*QbbLI&Sd64hiF4$mc1L%(i$YS"D%&bZ !j-LE#la 2*,QQf#*J)FDJaYcL"Qe!UVlI&MaLRLȍh6KbRb#)YR3F%B2)mtd6V(#Qm&L4E-Ѱj6##&dL#Y&iɫ$RKH,LRҙ2fɰX4iVʑTVaM"5i*K3k%Yl)Y)6$mlXKIPQИ(ڣhѲ[DђF#*X,3hm#*#U!J	E (M6b6HZBm$emCbBe2fR1IT֕F)YMfڦZIjEB2ٓ$Y5d-f eF2$`b#&A1jJJ2bM4T4R#%K4E YQ$(RQcA"Jh4Z,ZjK b4Љ*2c#&0`,lF4bLRBS	%(̪	MIHR6fѱE4#Ad١ɨ*2F(&M,%cE4ZMhBK41mc@ LdI,%IQEJiQ)66i6QFd4&5YiJ3`DTѶhb#%i6JbXjj5(4i$B"J ,	Y6JB6-%,Ym	Ū$BeDE6#*0,1-iJe$U25EDQJELZ6&e#*0R*XDZڂ%JFLlJbCY,Z%2ʋ	H$F#*ɀdٓ)R3bb)TUY"3I5)KbI4Y QI&V)$je1i5 QTXR#*dfDƕ"(,͙b4mRT[C* Q#*662LiFƬd֤Q!ef62ZTFXĤeFKclY&LQhH[m4&XY41heI6	%Xش[F5Me[SSlmQP26cYS5l*(I$I$5Z,T?	˽3#&dgROIL"nbQguR7#&Qm 3sD`0aAѹ႒AǶ28{4Km3^mm$(R*JP}v23Z(woDS#|%vu%0;uxQ#{HcQ#*lW'\<'yoo{9g]Y)7ߵUw\йd3V\hyU<e3zni(,IRxszo)9F&åj=ҏ;_&w\mبm#*Eggfn{[bU,#%r>l+A`,7o#&qxb.Qű|{#*oO/1Kqss&B-*/_D,[mf90Q irc\7q)5L%Q`e-#*5Fosdz3U,AU&N@n#"LdwJ9]ռ{F-=]vW:]:ʵb_myѯLx<#*J)0-`⡄0ĭPkPM9`ڜt+PZ6a'J]~?õ[iqb/K&"8ݣqd`Z(hU话5-+toPh(XB2'4fh]52I(XdJJee,#&}؇|Da$-Cf8^[ٮ-.+Z7w]bEP1_w]b~&=xcDv7E<G1~IxMьk47ͱX|e#*1A۴QͮVP~׭ECDIIs^vt}P'0pYJKIC$Q|.HQQuQ@_jk[[bd7TK~gaD1ejPQƃMly*Db+mfy2bzb,"2Ύ[=w9wkƾZP#*`#%)x<|RRiWnH߿^yջ*Hv5B-$:ǣ%]4dItUN(fP-#*jO4DxkrO#*WJ(O|9ADmpj}|Kwb$슋ɠѮԥb))#J_Kߵ1#&Ba":tMWFWFgR#*riG16۷Ka31i |pII.DaȘkS[inndM4pz=W>5sޢa(S~W#&Cdv>%<Lܞ$y]'ӕ<B@6`O$5{W)-&X}ځnđ(_֓	YӶ}#e+7,arH&?JNׅҩ;#l34*xa):oVBrJT-(Ib+?}-7Bn~v|qͧ{+~0;dC}D{&b0޴{EvC ې}سRb-LTcؔ'jͶ%=v>teg~c*$t@tN&FNqZ̷j.6Q`?NfTp[sAM/F[#*a,X2wGJ#&l!秮r>W#I POΆن,FɦF4޲MCE؂q;Wd@Ւ(K$oN 2SP@Oŷ#*h<1HH̣~%Jdm, UR.ܦȠdv:gON׈`$U䢋zktEQ#*ebD좁"'B˗дB>_zz{˺WީXeIEZAH]TT/>߿6rw3U~dg㶿tOھup:ҰkA5~Z媺u6#&d[agIOpI9R}i0М.%V#&L\=0!GA_R#*Y]Tqe%e+sEv1VR,)8Uz1gI<>[#¨2׮0.Ivt/Qe7e1=<Fn:9J0ളԟCe,~nЙJzNO]M1H3ѡW#VX8ơ[}$B75wP[oZC6JGTұj׍-)4ߍXULyܱ;+5ti#*wxwxŅCBĝDýc9֚rh#*h]DaKR fL@swм#*[E$$?+MuP~lf/#&	>G@K)WEŠ!!haKtGS8t2I.\85karws1by>5>.YS!K2_C#*EXUfEJWa0V,Pܡ#%HJrʞ	A[WF{e/9"C^\(MVH1^ iwdr+on`{)^ka6OaO/~w]%2LO9rRUn:YaGNO}=a1Rcsw?TȾ1(,yd'I>^Z2Rizߧ:-5Ԍ=A" tQmL#gmڇ]'4Xk]݆5t׈JaQxvWx!.t={PQ]*!yu<7Ăl:1G88ӑKlGaspV$OGic*&>g2Ĕ4[Oǜ{<<GמK>y#&{ɏ>xdG{2FڴH$=3|2PJU!|,:T0;tfk[7ifYЛT+R߄){\%#%I'd7OEƻ܂2/oZwԭwS	Z-}<M(fHŊ{RPnUsr~N&<*!brg,kRޯTfb"?鳉8=uCJ>ggy:yC2!%؉>sL|*D2VlNZj? vf2MsgCdt1H֓s25Fdn_gMEg=:sO~(mKYb؁Ҋ}u&<ho}ikDu9ucۜL%j7:l =Wj:D#*j-/ՠowrBnl.TfX`e.1bUVkS1Wd&W9tB6_#*ta:&Z{@{^{#&aUO%ak,e6&YÌ?91nk04ٻey낌w|>OGHѰ9ngFu	Gcެ32A2<5ZkHNjDuqy5+b>p"35%#*U#x*LSRxwnoȥUܝub2x&x߿&oKOiRIy_vXG۸ǫl3QP+mۖB#*#*#gtwG;Td{rQ_mc/}░b$3W&f>Hl n[&{s$y[Q*S&Mh7}3:7]%1+76P}!;)uGsᇽeu@OXdw_RiCqJX/?n#&caL߂4#\ª_޾91Ej[F`>GC`2F_Tek2qoK'qZy.{楞T{6UvTβBM;Ys෍CUv٭ 8r]w3H'|yNvO!LVZޝ5)df.Pjrojp+6Eu/@Jm)X*{pIE0tBL.y?\n>&4VPӑNz'Ai6Gddg+f[M=7F;o7wywD:\q&ڦ-hb.Pq#&v^>I]Y=ĸr*3d5t{ȇMRAKM(	X1#%$'>E;i{׌uJJAc8!6a_U4gjly-Zֆ[[Ax*wIq!H#&MHۆXۻ<Rq\#uQ[Bs1:mY-b,`Ф#%P@]F:~׌hƬ8g;6s]}?ZΧ:3.t|d3eTxVxl5o)""0>4	Q+#%G46s-$Ll9ſ]1S måǊ	Yzyq\|"TFT9htNq#*\7;/gx_{u_Hkpq*.}770=096r*P-3Ց[bHn~^8R}+o>:/wJڸDQ5>4&$<^(g۔MFPe`#&2PO(04ȠyŨrSj9Zud]L(4У=<#&yW}ʕT7	kf#䊥!b%Z{7rx3l]r#3JȔsi#8!FѻMÒB&8#*^w~ǝ-QJJ#*;d	n58l#%M7~`KH((	z	 H1N0qQt$9j,")o+Hqv:#%#Ft,bϟ9yP#%W)硹(0V6&E9~cg78u`H2#%lmh[vgO=1m=ͦP'nGh-j#*Bs]?mƭqx31sL~¡2IzLVKE2P>O HOjϳ-H{Jla#%D̖h[&-tzBP(, XAEbG_w k0K}4֡Z#</b8CΠD"h#%>(YFL}SN?>DA#&P#8Qcj&US,V	憵L0yk%mٮd#%4",}#pϚi?ݡ7=z꿇RJuN9a4~v~#%#&X6юaE:;̈́d<5 (m`sŜ}_}0ٿFY\Xk_>ȁ#*!15}h	,ha#*yKX+هT^xoyy=`d\<EYG?H*JՓa{ͱrTI$Wp?u9WP47^?odϵ6،<lᙼJZǟL;hZqӲaW%rJSzHHôE]a$m]hOFhQl""^a'tOȼ*EN?rPޞ|q`ak֛;V&g*<bهOwEF0P<('	Mck`,d>nK{b0۟ag"ǫ0AMOb(m9N.5NiݬPI6y¾]P[?cR4~ZhE#%RRY_ES	G۔Á*!4VC	*@?t#%%]՟iV#%O#*5o-\G{.ܘl"^."#ۜS{rݗ{LeK}·uN(ċc*pߊ)0۵j+ZcCL+{ [#UIvkD1ty׻QbDochmmC"I7)HTu{k?}{N6WMb;:7F|(wf>}|c+kmin@҂LUCg:8ڡ{"Xt6qVL( S07ŐĆbKg#&k eQq@Pm\#&,dq3u/K\c6!]JH&\oX5ɸC#&n$qE}v'c:`]v0Ne9N|h(1XP}S=RM}EO赦CV%UAjEYfI;kw~1gu:K6<<;;6/rcql.ӘnVp=LHx#B^Pml"E;#&DP_5Eėpd#*µTDZZN9IfT՜k"uA:Yr8PG_'B5#n3ᐤq(Vvxwpi\~QV%UWS#v=9X^э$qr6hgmnlv|5m].m2M{r:'⬹Օr=qo[uL7v6`_d-Zd>Ig/3<oaƵ~ǖsQ$#Pkg2o0;E=EqTWlS##9ۓ	3$	Bcڰ!^g騮/JMLQڐt#jf"g&qZ!q)A0Hm#%8B#&8X5wiΣUEѡ6j5:=U'$gEb>v#&ʢ6ZvhfE$.(1[s1WyG@e98^gTaSmab,H3xɐ#*x!1(iNZMs;jup'F0!:DWlXc&[Roj]v'_IK.WY;r4\}53&kϷjcVњq6-czݩC|$̡KlhE&☦:+;`.T4nL<34f0ZCMCƌaXF80#&V#*\#*tm^%P1(Kk!A)/ۉbOWZy 5dQ5^s#&m0ѓ%,HL`m	E(",",害ǫt3I=/K|=fN:>w{/ncA#&z"RrkۨU6	.ޟϟZlH!HG{a QIؘ9,C?wۍY#&s.{!)ئLH״K&WJ60*ehҒnH7qdasDmd۪KU3{,hY	m8i|_I"ZvTkl^>֠A"A'm#&*,TP.#*~֪t8ԌajO#&@ԙϝIv:t#&__CY)	AGKh٦P"g-b		5%u#grcOMA!;WDJt<1#׬'*aJ+#@~'!tf$ϧK	jQ}\R<Ml"PezEk#vMoqRky|/J?Y#**zcB҄89^nbE4(dY$Ҕ:61F:E-eI`&"B,gPh`!tl&FA06	28P&J4iՅ#&<f$1Z`Ivays%ZYW!<V9?E&jjlzș:L<H'lk0?wtgxg(qU[%q"u"RtG#&8S0?5Li&/3a7[%jYX)Ȩz(`َNɼrNTkX6{bYu,aGbET6_r_LNAC%6?bkx(QKSo;({KO{z|fT"e[&smPQ}0iI	v|S2q!X<cU~aZ:>iedazl"LXRq)#&Hh/MfV69rtƖ8Tݮq/|w`^e0tfۗ&U+5P1zxk.o;0D*'\>pRsԉSOv(Ñ?{#*YhD[f+yO>#%n$#%A+T"Pʭ/QW5o9|1zCZCӝLd#.eEDc	nUTkQDj"Q"@I6*Al@ApboIӁs8)|5xbCte/XR{a	~Dcg>a屆v,*2۩#&o_%NE':d%Izj4h4|C7w7J*TP|Azʹ9paln>}A#%螱UYEһȜCxL)qh17R&JRĤeLQPRv'|xV˥`R9a/Iaǧݴd#%-zƭ߆uLld>vs?\pf%yaTݯa}n=Ul5afA{bL\I#%-<G)<PX?t;=-dCƼLYT2n˅jv <n$E\#&(Q1[3x6⭨el-j	{]mXF~۟g%+9:>+g:#*8޷[}؋#*gh"%W2b6T(;Ѧ_ԯc$	U(DvDq]/E} M%\RiCƆ**PB,/1`[(2UPJJA={nu|v8j(m	Aktfo-XG(qقS%![)MRh2  S9ίLjM&CmL[#%ʪa)Vf>H2ҭuLl06Fԁnxή,!*wn2KcKڰ}_?/&Pjb\閸@\!R:#*m<I<xi	Tr	Hˊ&yN{#&Au<;tAKsdI#&N_*}T(D\oXdQF	闰1K~7?^S03fk!VO4`'Ja-?ͻ@R[:b;DKO`>+z:qˀ5#Hd߭s	r4t{EVK#%L}Ի풼>gqxǧ@r 2Eַs:)Z&.3Zpj+JбQk/ߵt=ecÿ;cmTY'~yei8p~$6IU@ꞟ-w$?UQ|/FoZRJ[#%l9h7I&kYt9Fݷ~N.w&N41A|i&KoǋR#&U?d9!ѷ\zD?1	}_RLHr,91	<ƀɋab/~#%):HZS)CyM"NsZ@c}I#cG6ч쓬hVf@~޵3ra#&;bˑfB	(5bFi(swCw\KF6bڑB	~|((/`-#%jݰ8'u2z L&%4	Bly=$Cj^Aao=a)6/{T#%)+Ь1!rĀ,yM>Un~:eWIEn;x֟ڬ?k#%Y#*gk^	p%Q6R)oLR]!HFA|X	nz!r$#*	 afeFFcu<ͳX/JU>\m7UTq͇_Ӥ,*%IK)񣭾/0sP$[tMdq.jQe<4<n	(=w2[|x]xboٟCc.2BpwwӯH8۫Mm7uyi_&^;͚OOG򟯾A} ~AG^s_rUznÂb2':[yӶo{f?Pږ2Ѻ;ZjwJK=_1t`,|uƊ.(o]}Su*@WKS}RU3ǡG	O=Eb<O-s]-";\LUcwuS¾Z$+{D^˗_VM9H|8sCw#*+T5$	@^V"FoT,3s1͞7\X룒_{Ϝ&3Qv.c}N#Qcw^U텡U+#*ogmMFD#%՛.FO+@F^s}膘z9Dpxj9?zLӒ|-ڽzw{q%Y0g`O#*zuYWQ	#g&4[\3Xoq~;ÊRB],Y۵Z9.@3_gz.\A_w~rJEare^E?>G~_ۖa6#%q#*z7/sgDoPcTvQXgnYۂR]soNyW@z"hEH#%QQ\peݖ/	sutT6׷||:K3iyMK:6ǻN.x/Z?E,?~Q?oÜxqw([m/vzPċh)g}_xч[156	GSEz#zՒz嘠uPb#*߸P~sJSs]#*=z^#*~4@~T!?<:i@H"j#&Ӏ﷋VSzn<G/Yb[qv|u.|'%GhDPH;G|K~wy&-o,@{nG#%Qҡ8$T#*j\F3JIM@(WxsRh@2	I+V^~#E1|Zaa*}x*DGC'Ӯ7=HoWTL8RV눴K8KP 4Mv9e:}!ӫzuMʎmѤ#*Ϩ|(?\䔨VP㭛ptY4DvXkYyg&WxsWUqN^?53!JJ+<o?p8;ŶFd,CTF$馟hZ*	?1#&i*pWm -ݢtc}KE\2]Oc~C/l^$ؼi#&~OԏPEiqW>8#bhmE@?FVx{onHAP^;tT'ꢙҜAOϛ8=N#*>8g/th7iɢ(Gu7lr>é6P޺gAtO.$əem_o>hG%ݾQ/smM7عC?w6k#&LiA9Jׄ^jusX')]lL0vP [_vyWJLDwy݀u"d( )䆪:@wv_6^>Z㇁9ֵs#emJzb T<keF=0 XGG5(D #D68dXH*J7R&([}֤On	3pfLvihd,ƃMIcn1voiF,D0))'q8&dG)T'MZ/uY?g㲃R}ZrZ^3cIX`},hqE3C0w䙋(R`)EBzm7&d?6*ޞlߏ΁/ޘ6Xϒ,)pe>cUY#*<9#&aQGxv<rCi9-ˢ3j.weMTQo7{atzD#&xJ;|ߏ޾ysQ+B<~{#%Uk{#*:>\*ݷs\UȒt;\3ܮûw_\F9O5vr˲ڿl1Cʍc8:b~T1"ּS14f܅&բEm)JvO+FCC(kJdRWn0ģ#*LㆪPP#*Qi"FeHy;'+oZs#&em+^'eӅ]tn0c]R0RMFDI8vUt)lc"1̉f*H|AysBg#&C*`@WxY{5CcCH**q=XVdKZUVME\Q(3J0hG]dc<fr,)I^aV+dIaE0W'DAth,jh_|\)b0.#bt:57ȝgO	ήki#&v9;!Z;ٚ+mu`˩QO ץqQJNIBx{<i@y<ֿ9}0Vs/=<,$+/?Ѩ邊b)x^b8SN~dz-wY瑹4~.\RI,ʪgRa.o`xp6r/@^M']e0x؈K$8/*t=mpg͟NNã9/h}&HF"v%"ɗ9HW)}鉀gps6^_Q]e~CeTWնmFHGb;	]3#%,**bT0jjg0e_rU,98{`  PPIJ0&}J"IXjA$ҰΗfr%8'1kPC7AfcDD'c7!Z#E3\eAT`+Q`|KĽxXk1qVգ8a֑G_ٱɶZ~20ՂS1vȊdLfK.姣CU3VVfk|L%Al2HPLKأ-:Ct8LmX6jndP8m9X㷞kӽKt/޺	(IxZnw#%n\x0#*T9to`t5jF1J	@X"e2q~6߮:Jik`9h9AFϫG!2㢚SV{i/9Vwt'@'@pZ#&De*#&O#* `Ӎ#ܒB't@HaP0=Nk4	$+A5;oIT:憰{CF=7ybځ#*w;B- BպZ # tP7".[ήUYS_vЬdq!PM)(iz@#%2菣n)O5Ի!,vxK'SRah$\R7P|Lw8La` Pd"#L1΄x	0d`ۄUFU:#(٤q`?Bňj#&-c[eXk;l86hd͸="f3ILH1(FڢFR٪,z#*|iW2#*|@EhwȃqƷ&9$8`f#%p.3m>vh.(uZV&hL1xrf;9N-"`j0t<LF4GLzyw_*8TaaߒLFosP`.m|X[xҭA7G{xv4ڿRqAxFJ8acx hv-:"OsQ&S0HǼA\YLIx؛2.Ӯu414p#zT[G7K-Vm%TU-]viex@gDq36kܼ6ŃR񐩈rݓp#*d07,PHRjYߑ|5Fh7P<N=9v$a53RdHU,P&#&#*}`GM0X!b܆%l/"Vƴ;$tzsYḺpa=Z<Q;f}haCۜMf,#&(_V6l󝵬d1Гaw&Ѵp['L^V+U]ʨ;DYb~*;me% F Goʍ(|7-tTG%a"Gtdn];Oxj:'HK%oNnqY峺m+`*e!3+Xq~ˈ<&)jiF+k0xƩtf,3M1iEXpG$#&KUdo`>c\c$6,')49x}xܫъȷr6ӈ?q`\=R<tin^:ؚ!DDT9"(wǻ#-/}+or\ibT+c'-D	,DSZS2nBjm*P\WQU\,X5KᲮn//#|R8HRBa'hysBxGK3eN.6(TtQg<%{Τ&&HrS._~}G{N-q7h'zW]i0"nYbdCs?ԍl:{Lҭ_\jl嶉Zl՞2`g91վ|m(룄Ӿv[7:7F29tg#&Y1tpNӶ]O|fa>jgp`D؈w#*aF.Lz!4t3Y%J7qDX[һQ:ץ$	0;MIkXj	LaCfN1Gq;N[p70xj)uۯe@rYK!}hú҆btk8*A3ͥb㐇#Y81]o/yu$5M#&^/F~ p|DH` _dgss#%y~sByP;R+vۃW:ןlv:#&;t䦪f,$2r5hCPhJM!umg,FC_	]o`q#*.+k,	[g`禛R˂#*CF#%F#tJ<&;QRVb`Am+DQNKZ"##%\9]#&RjB18/v>UN/h\Ct(8#+d)Y56*1zѾ>?G4rFU]u78#*KS*У&b|ἳvK.܅(`aLSnmLaQZ(v+=w+;O]rLDvM1C)bǌ;_I!RBAepm=Osc5AQ3:ey#&v}d'Z--ޟQr݌|:rOŒ>tPMG8xLjo<3Z˲(q=νA咒zjYG`qa!=tljFKٻďuy~z~HKN\Ϝ8l5OOLtwsANfۯ&Zs0v7Kxc>֜t{d&XC-_Rڨ#*X2?h?.1!#%#&|madZlZ3]&ԖL:8~Mh<G9u飢T'H:IrAF WED3DE[^bLz#&)X,Knb+].a]Pz'tBw#*^A=tw"?nyݧ#&d#&1ELF><WӾ HlxN㒂L@$}8d_,*՞;M:=#*ۏ^T#&LLެU偁ݿcdz:֠he;Jz>kʔ_[m0כOhwnGi5;|/<8v#*_efU<((&l㋪Q#*r{T<k.n3P$R\d,"էk~q wo3Yz$k.#JhU#&N3mMNQqc`EAdem*$f2*lemZ	43	MEk^oj8V\ۻvQž3l-bCWJ>Wtkt"N#&Bl$->rXm'l7=2=K0\*BEL;mנt{4{knplxւPB."ѕ.wm#8㥲|#&`X<Cǋ/gq\˦g8dtcxzb4o}4ͶrojqqϱŚmi'іiIz*8i1%>3*H܅;ކԏE9'>MeרhxEa@Clrd|<Fim	XAC*Yu<Bs}e/`^ska9Tl&Q@5`0]:WΪlR9)C_%#2Z,lzo[sv7hJl'ޢ⢢8|7̒Kq.uiѹzL>2jǗ{c]f	M!['*UTxm`#vY#*ڇ]Q2-#&C<kݚiNrB~kc;~2Jf|D,Kj#&!}Z1Υ!gltg (0gob17|pHԈHtt}3EDq#;SSg?ek` IƢU!CʉGF5o'#%kuC{M{ElzBiX@qQX#)δ;˄mhR^#&ő+Wg~e])38T"UFt6"i>ܻOyrpiC..WNkiКPv4n-B!4#&ŎKs8Q85gJcwƗ&cږcqtzqs*&&:XW>c\J0#&AUd쫤Ia<.za,s6ۤnpB.Q@Q!Յ`*s}r]JnuwPTMTy{gE(u(9ͷz[ԃY#&4k8[#Ĺů.!)}|D4[Urmqy^퓕a7r<8<#&qq.2"$4w:QJNKSÖXg$SE覍~tRdS#&,-"]=J#&(+Cw3ۯDX@ZV\"(`ݝElŖA)X.UyʁsTQgtGP0۩R·(]&)cب.ELH5N7NɴP)vBb5H-J)=yyn<6~jR!dٲXUs8<k{6Oy|Hy{;~/LT>vvҊ9&H#*[>:eNkrE((?PEq#&t#mҼ^csz[p`XM7)F(T\bm07fʺ*"~N˯ox2bS5ˣݢ1n'\Q-5YЪ,Nĵ.+	W(fY:O9GH0Rպ4:#&3(TzbWSyq58_K(s#*03-Cdh.lƹ¼S5i_]caS;y{&hQ@¸ׅ&\+<y;,eU#*D$\u_)Rt#e8HMu}s_vhRE&5 ?I{U>#P#fo>x5oG8;dڶ6*B:HH,\ϐ-N#H8h\(z3Y`5]BVe#)qz7bQa-$U<u}9޹è^4K>SrHaod*ANvu]7CdQHz;an6˨p#&>9_V4Q	;;aD!aj!<MܱkjaZ$3#&htnNQ)Mv"UbBQvky'ذ]c;J?Gnjd=MuM_Ly^޵=3cJR/#*$`%HcY]CKy#*%φv {0uvؾ;?>z_^{?#0ׁeS|Sw\ۑj?״Ruo;~ZH#*\)WW0gyU*Zv:^:/냤!x7aĿ1Lh6,CҪ+%sT_v׽>9ED3g£%N1`vqjEbNC?-7}(:h[>M4GgpMcq;ٜm#&xO׳_L#&WVCL$LT!KI:Rh>j~Z^%*uXzl(#*{#&Y{̨wb7["V2{YgN+;.1BBLpսz̋GEGO;RS3u@߲c*f3-F͵#-FoSIPSWxϾ71/v?TJAjN!WH""FhԲ(!^5W	v0ڕtG;!|h~WRbS!}֠&?ϐ(\T[-a!@iKluȨP-1s61sgm^jm0I^)#*QH+dE_#%OaZE^zx*18D#*VEtH"fFqP+ƙ.Α9\FNm6T4-Snk !ys	Sp~KA`9C4M)*^ʾgEHЧ:3*1h#%DvEbf-uN*Y#*gPluP;ٝow.&&K {gX}H5ogzusg.®-]'*7=j}!c{IE:#*U';ձcG3=3AB;uz}+<Egm&(5PCQPHMU@궷PRܮ?AΜTLI"gBdwFtYK8.0B_P0 2hѳ鶉92mCpH(EVYZ8a23Y+L^b#*5ߖ'@`#%?X!%tVs<:_O3ZPڗie:Uzitò K#*̄)T&ЫDhS)&ģvTCB;jt%Z]3ŅJVPZ#*J#*˜(·ءDJ8|rm\\ەU	/>dM7,h1VG˞mˆoLOL]ʾ<9uO{EMk)^p:.Sat]ޙ|wze4tmFbty.OErlf;#ǉ NډAdUuQ3UYim2/o8~8&S9hE#&\w*;Yc5vw|v%{ScV*A" ozӂsh_	Jp8:þ[[ʎӶ=pyg ꦙǏX631G5SDv@X629;D#* {f6wf||mÿ!sdh$&FKv,x+Ψ#SƍW;Dg#^	T;>JnT,-d@p+,y0+k΂F!O'j"-[okN5{8GctZٱ^7S}sTk{B+<t5gf{7|:pt:8}=:ۈKE-rMو5t$@PHL5ȄnIrĸ\<REۡul3[94o-jo=k=Nć`!%z<Iw,`䘚wf9Anv\?㞇&0J!c5^oDؖqOV#*fMK4kzCqqm7otΫ^Pq?O3}4T(w=/^ΘJ>jڂ#%:& X6V-ӡ{ ģ :O3#%|_uQU<f뀩+g#?_szέh[M#%	;OTT戊߭dFC}-leaEKӷbлxւ?}Aêl[r9Έ1KzS]Bt'L#*@"	qO[QQÞ+J@8zI33:hdWL)Gs>K(W6,i@^*!|,l.	N#&*1q2ޝ?OEtԟQxǉ19b^3r=T}n n+o媟ĝ>?[~#*|OI?~p{6p?Ϝcd	?gCRoxE:a!woN_<$ZpCb|RkEܶ$_E#bR*Fd	I#*)3=ӝ9O"FlqKn.72G;NrN$Y"*	@$}g_2Ss5|G;Ulj'2?0ٰ}VV * x1vaDh6)l`jHUV`>#%Q$6A)6k͇!5ĥQ]&9N#*|"5#%,#%}Uc{,r1#*#*r9[pjk}##&F0dP7jt*Sλy#&pB5$9KC)ba|W$Is#*EV)^{q"~SCTCIW#%z@;HTl+#*̷#*hnϝ^%\īB#&RTFR}T#%CϮ<2OCS:5#*@֭e:Ƣ'f?>HCըK>8٩:P<#&Hhjȡ5T懭O¢:5:NQ=:@1`#*[(,fO)s2$ 䚊#&BP;5QB4B?o翿amw#%vFPkmT7޽<*rOghYq!{A#>j@+ciGFƕ"c[>Royїd~;tG>"L+Ry!H oS#&dҚN3P6o||r:/oq bm҇u'AϚ?~Fw']B#%,wꣳ{0vN=#&lp缔\Ƴ9O](^;jVl{(T6"[YQGO6rаE{\#%jYO.6'+8w]Sq8;L`M}&-F	#%85jdHpLd᚛#%rS'X_˼BL!һX$f(tPq~[5hl5@n(/קY&Og~@P1yI;v9-z#*cgSr+xB<HHv<|?^o;/A	$ܿfXd)jMBV˭b!i4CV#%(h1	C*S2^(z_6iF\E'aqD!n}+}0q~f>=pqI&U#^e`0udPWݒ"Lir^ TiȾ*d+WOR "Ó0&VÅٺh:cٙ8i#&R]-Lx93ï*3$t#{W3<C[fLzH;_"kcE[bEk$~Sj+#C=+1mm}҈xst@!b#*sA6Hd3_T6\K0|Gg3^מצ:̢:Ӈ"[όLS}/e{Mp1CnPIAgv	q!";#&k$f#&eT=!V2r&0{g0-7̈́ó@@gfz;e[o2k	_ R0ska?ሊ\Uo/( 3F0+kc]ow8DLQC_-w pË#&4cvq.K11s\otʀ!**rGX4b!X/螟7kneE&⒛DKK]ܴ׊x6R'qpvѕPv2&.ܩ~W26	l`,d(gzt.!洋u04K=<ƕIJS_h"a8BZ7]샚j|} <[S#q=X|F$N=AE8FX(Ut7NCюZi"4Pgyظv0>P#0m856cUo2\EXk/R3Y5.|E#*0(Eؙ6Y=CRK_Ï^p?d|H]^bha(0.#*Fa.S\_"bc#iGrSlgCdA1Hl萮h6>=#U4x/?u#&Kt*M:3Z"UoHq6Q:M.W~}~leًP$aVKy]/1&t8G}#muvF-bDRv(	->M/rgvx4!3볇f.TOa7$F(ET#&6ڤ`#*i1P%b(i(G>1y#*gLa7^#&rnyn+_AJbn diчU΍3r3Oמ2recG1|05e7#쁮ФquJ @9~W%UwciM˗Nݫ<3\I%#%_㸌@TL#%*ôj*#&	)@_¯&=#%'N^|,fG#p|#0{-}je[{o캴#%A}ɾ,'xK@唷<Ւiec#&oyArxҧ8>q)Vv:@\#*.$ҪN3mMdrcᛐb|DM=dGh}s硺_kK,T(J!qTSr#&#'U:q?IlĮ[nB>R2pi/Գ:&(:C$8n`/U((J6'XtRj7fRNh{3n$gFz6R T!CZҨ?ݑY};ݽ%X-L?5|7xaKPV%Adf.@䁃B=DL`H!}mr,ȅ6R87#%`  p!`{EdbAIgrF"0ȡ.P{l-ۋ@Δ(XEa	l8$i"ȴ2=:V (%m$Oݣ|ꂿ8`BҘX~yeoWpϬ7]^sɞB>~kÙf'i!f<<78kx+{{y_Is<3MA/|qPFQ#%vD~e[F<߉}JV\yq*7r灗BsU2WH^\h=ҳO&YGDxN5bGr[ƉqqPa)U/s.x3GUF(z48|?5vط][먙.iHۇ}Kx?kS\CƮalнkJI!`1b`uZ1:4Y"Q5ҏww'*3A5'<#\C9m	"W`]dHARA&&R@#<!J;@5CEݻ>ހz+$F@H-D٬Q˝.z#%jP$!n,,Fb/r3co&ަlOŤLZugx(}M4%B[l4~-6r:WƫM6t$5:p7{bpH$<ѻ`1ƞv;8Q;MA8rydMpq8F*P2JcZYm%[ZVUAdD@{&e-%\BDڇ!.n͞u-qT 4x$0v.7pRf5z8rvu6:$XMuYK{pIǓaMQX6MXk|Ci.	y%3W>"?@E_0E("w#&=h:+-JYIIoLU0Rd:P%ҵ˛_&Gflci#*F5 P#*)Q!-EmHcq"*PB;9ύJ:N#%#*QZk&<d9Xùo*YSq9~Ċ#*2D!!{r6^iW!]΢BrŘizj\G0{\ChlS1$w#&/bD߮5D(`4GJSJ`#%yʡ}ݹ;&&zxvc,<6xj3zFG_X-opSuW<~#*##*AXWnWR=wsjQ6ʥ#*ō|ɼRB	XǸ:|ح7'LQo+(I#*tm	h%,32 d@9[7#&UR[ӡIi3C#&V_U=7/+ON#̻!2"?Mu^8n+dBjơjh~ AV F64$iqkzc5s5QFX+1#*,ZHqn{_#*u\2pҎU_Pͷ[.!$Ijɡ#&a\f͙#&#% 4ugݻgy^7]q=O7YTJ46F(]/R[txQ֏#ݱ6HD488y `P!^L EwHK|wT;[]mu	 o8 4\ۮ́;gZN6WnFcPnFIt<{9L՜,\މJW0M꿥+CJew'#*/iSn>yD !c٬5g5_ysHcr,37¢h;}QyM݁,z=_#G&Mk_9bvM^{~6m=CNL`xjȤF CҖn&IWu"O^E5'#*&nB"CcW`~#&&Z'8p*$ yB8z314fs]*RF!S']!/ɿ>h{Of#ﱸSt3L`@ۙu+xXz*ʜAUX#&ӧfD_8憻3g^^ĤƵY-i-靀ٗ7JЙpNFDw#پ(@1DCm(evU5##&)׃g2q|ZFbhd=65}k\wKӾZ烉RTֆt#&'hCս7nxvuZqއHN,!!6TꧪE't(s	|HWK#&ty78/_Zc37$m1%\8-){ɄO8¡PYasw@`j@Z/M/DOP#%|8w~{L=~ +HcΌEushE#%#%/2\xQn> 	{#*qȚk`<g}d#%G<<IPa߯*;?KoAbj=8Zbo#錴g|h#*5#%'{9o+13QEldFHLiL}UmpIˇՖ?g/#*CsJL)T]pR7CD#&N)$xenQElKJ"(?/?/!]4T|*ueZUOTS?߄9jles3mԃƁ@ 	÷UjV/<}µɾR5CʈL"BM'Y 9A:Y˜s!Lxq|<msUih隭q#I"!#%^(RС(ZJ_?^0IWUcCۛev瀴HHVT;30ԉl#&@}t	@	F<	.`$G&ᢾYO8'y{7M"uYe[?sigu" 㦻m3rnNצvkp>Zɷ9M+,:LF9+0=ݍ	!@SF(K>ߗQTǿ'gD>0aͥdA$b2%B-|-<oUeWHT_aחc,{ia?5XDC.<~{W|ƄJ$A<XS '	¿/? 4WWc(2oOPԎsO!ǟxG&'#*HWU}ޚ>}5fW'QLdD@ xFE%P3	p|ޯ¡WaEt,9z0$CAwܻݘ#%G΅3Dguutȩ#&G;780˓ū,Ow)Mn\j}4#X>se#%/v˃a:	?qxw!sEzv'HwJÓ2clsM#/:x-A}#_B-L1SHZ}75> x8ti')3ECŇI^3hoixIԁDӁ@KӫDS@~}ەs1Gя f8jaH&PJaY%>|(BrI-BQEfPAzX{B'sfqU|mCq}%mO1[UW%JEoRGKSK.x*#}nn7	'Nb"1k10	U0qQ+12D5T"-b#*:n3|Ud-^-8tpdǷGם|EYU#*f	HCrLH~ /2h.#&^m[Col3j5>#*PxmwtVW#&~mp|#nie#)[|Ƚ:2Or]}#%#%r{5vOQge.x>@(#o@#*;|B:J*6@xQU9?=z#%`c<lv_#%t}E#Qck#&#%#*~HYzn"5j[+gFwX]#&=p"F~Nbpdgʕ~>v{Yџԏov`:}i+\zٚ7Nlo#&:8l_JH5秄IH޵xxyUnj.kb³fu>A]yd.ۘ7WTWg.=PН}TK)Zy?)qS\b_fuݑ<cĹǝf6\#:9='\> 4&ĝW&	媙[,Ux#*hqH0`#s=1tRf`T@woR3n)i ,jńZAH?:X+U/-ަMG6a8B6%cOlAxo:odJį~GM+;j#*ˊ%oUxӧN~gԇ+ںIUEb1ުdsX	tNyzg^u4#4SsE;8i?}oߖ,21t!޷=MUUvz7<ڢEz{l<3^y!	(9۱A>NxOc5Ѩ/}ƹk9;.S|*X#%,iek#*4ǏXuÂb]#R~ECMQ.]"E'#3yB<"O#/s]O>X}Q\kשiwYř}<TSYA#&m49n$1e)~ɋúqX6pa9>?눑n0И!{nnIE߻1850ȃx*_&<U>ouf;vPfĠ	a:#%uE+E4x!훏iODBgHSTɠoQR~,Xsd	U6E̅_JxńJtZSʶx&퇉!SQKZȌys+~=Qq9|c3[x'q#*Gr>~\#&̱zwg?ǵ#&^Q%Ϻ_߮+~Vn=(kQ*zyopkk2-!3R^PgRyIe=I\B+v=<G64zSSZ`QC$E6M@K[~cyW1R|0+Tgu}e+JŞ#&t6pD	{c9^F31i˕iNVk;smP#&ꬸQ3*/z XdܢS*Fx^V+>7ğA0Z-Thesr#*EYfc#%48"U$3#*WoοÆ>030u>1^4fXzHiA^gKuҢtSH=^j>J:oߣt|&^0gOv ;1\75dN:,ОҡWq9E#jK4uÞLR@2WTBÑ;"*xo|-{ܗlGf#&~honMPQ7e&za# ^o!I:xnj{`}ܞZ.З JV$ z@#*^0-Qd0!$M X6!#*x7QCYW"Cgg~o^Jv?H谉ggJg!EapG51O6_&0VBf<?ywS\?.ٰ01C(s4!#&5__!0 y@Oi)gQfKoq@!t<hx}ڔW_SM=(#*|w >zZKޭø:.A4og|QZ(4(b\1_(o95|ㇾUt~rd3~ro}3p>}rի@^A|D$" '?e_KrpzZ,yƾs~|yr%$ܼ(}C.F2OݮSWsr.=ߺilB&:GVRwp>BD?̨cDBڇ#&]܎Aj{t=gb:#%36?p#PnXrXL{,g~Gմ~ޝۻO:xZ0QqKZjդۼ$D܈ gќԵ62ě&ՇP#%Bpq!9FD|A?_ݫ#*r;<CrV@;r.EDr՗OTOaL_>fkSMw]:#&$w%U#c Az#%J#%	,=lJowM@)gG$g1Q+EDUQ#&f^zn!;7Mq6T{'"||E[{-k04ː#%q7{ɯEmLzP>:Qִ5e8=gIT+J*#xυZ#%tܔU5oBHЊZ5$K$#%#&ȍџŔGس7#&6+LvF#&uXعAfRgZv)c04]ǚS`_ _)0o;u	A(Jw>o=qymοlvĥʹl49{M@?2HJ2U=,k:6#&v fA_&a)QƙBcҺb9_zhwA/_:H4}oe3ZDWU`l,!x {Ab`6,QLÉ䮧	εf#MrZxcͧIOi܌ @!ZȔ^aɄ ߴw`yup2ޫ8mSnS]cN-ߣ8ƵIXL9<Ors2umuiE08CntVn"5RұK\7NSG+5Ì(,+*Ear,Vpoy/Y.9+\|#*u{}kVԈ<#0<≙>?eǲd<vS\n}Gnh[KYbb{R@#%75Xcbșm'Q}OQgWh<03X,3{:qh4FLjEYʻlmliy/7ך3iʁtF_yiG\mpב̕II6gwΰiUu^^B1߿wǯ77P4fu(:&YRčЂK)T`$3Sl,/b,Tp2BJYPe{݅@J4C^,lH2C%7atNro0J"RNӈqH)`%\yGA`k&6@@zTeq`QK#%#*pr˽,_}20џO>{KexQHT3GǶ84֏OʴwF֝ItS.6<c,!DOz;p]tD?% O˻9B:I	Z2Yl鍈lZd,][&U+z*V+VSE:#%u<1,d}r0KbُoÖVZXz~OG%(b8bnzGbJEp#6^֞/%n4@;?V#%l^GxnM*yo<]2RO"#%Miѧ-2"mU	_{S<Szk!m׭V~-Qd6¯qu&(Nt6XP'#&1;WVցri^an#*ݱ"(ld#&p.0_7s_O[#%GS8yzi)LD~zۄ_G{zAAz{SGAvtٿ}?LUpD|hCZ tQ#*ND˲9@+qh(b1[0}ӼX"1Mj:\Xav7:kgOqZ0Jnhq|\Wmvz/ROPVFŤO(5(mT3LX0"1մeѴI|[tf8_Fg@,;)jy]$H(H@V;Ũ_-z{\!SKtA@yuy֍ba6I5U{?b+F;B.&jnHЕKʮ9bo/`_o>$g7:	<qiEE8*(\`ks:1}d;n	RCm'㮏YKUYan`=n˞Qml%G5\;$]L+d7Ի{#W4uku΁=#)NF(	KClgsM]*p*A@qJ%WΆd$Uj>V!j#&R#%{eCm\9!kĞ~gVћ=yhhP	\A#ijW4/eLλڣ]Cԋ(dGGCA8"KƬ>W}iYO#&^/RcZ#&m{m^Ǝ+{grePTjʈs/e*YŶ"Hg_<l0h(xoB&iW*4Q/ Ղ(x4#&TɎZׄMs{{Ϳ=˝|ܘ_83nV .2X3:6DaЧn7uNf+4V#*#&v(1<[O;3bAQ/#uIEd릻76#*HRv)u	nQn0vaF3EBH1FAvXn=4^9g,Ozr)5o#&;`N8ʚO6oxE;2/=JhU"Aϕjl'Wj#*ЏRnu#*)#&]Xʠh)!d2qxCsFؠk\j_` HA3#*B%D?ur%5*AJ=D`U#%#*2ZiDDq0a?8GïtQ7I=֝s՘x}[HYcX_輽ťx{x[B5ql#&kg4O8MA,ګ#*#%;rv*":ʥ@JD88m}k$?(B#%@ý\5ba??	.%P;u+:#6%Vo ,#%w7zt챼avne훬#*GCcǢ0}SRdSq2ͤ+i>GͼX)BLcT%~*QMf	ku\zǴa#&/htg+fz+p/pwNfZG6s] I#&}8#*Ф4G/ѭ02O$=pA,Q~wM'L-ΓUi#*9iT`xaf/xJ[,Tk`/`#&*Q+Hz7|#*F#*|ǷϫsEv	$IUW\#*.\>C57TV>|e9`r9lBS_7"oE4#&0I#%#&@S	#&d"|6&b {$2fx#&)%db󦫍@$މrmhLѵF6ƠM-Kt4 $Yaqs\[iJ0@zd0"ÝD伎#&cb=cua+Lt-@8?rEBex^'OGeIPw`3Zoˀ\V@3f)?tSHe%L6մ%'ヰ;{m lu|40 C$z4S?:Z\u[nѹ#/93`(LH>Grs<i)X`{>*6@H;PRJ=8+)4"AVT&.`;Fq4`9f~E.Y[Fk!<u22_	*F&ԊFo3/ʪCp~)LNҠSY΁3Ln#&ڗ.O9Ւ|-<Ç GI"gx-WS(]4ҩUTuVyHN9]76#wX?z7<#&0pt# ܞv&0r6.pvr͗tK6Т3?RxO}|c#* O|N5V'e1^[J?߈ŦFL{=JO=^6Rp9##* qv?3yj^!ZNv}׹ueGVCICG>ڈHUg9;KX#%#&-,-;6DC.DC!ˎ#*DHJdGE6S018Ӵ6/9-EVBì1@U5<75+}gkY#&nsI;IŇ8Ƚdx	,6h"Bh)aIuђ1$Ex`Mq	#)PYY$2uCZ#%`cMc1])ĺ1hg$F"k_}}[}]\?]Q+UR.&1@aAdQDY6k_QA3_ϻJmB5oC"z"U8to+ѓfI$$)1S5Յ̚5e5e{ַ+0237#&:(m(h	Ֆ4hIJW]{>pJIH0BAS#%2LpO&:*Jv9||[tQ{м5D4yUwJ"nʸB)KBEgw>}j)߼7H#&hބdIqE/vؖMjH MiQlЪWf` RE&&I#&3B[\$I60<aD8uv^x	aB:tmo,YHMCcW>p0|Za,u> N(kDT~ӡR-8 I=mc/+`|m1&["xh}%eFY*G.}BmS~sٰB/;=U*JRDSLCn<[itdfkt)@T#&57.CH!mpZVQ-o]6b@6l*}Ϗ/	0fa5m躷@`)?F*oq"#%@ Q,8ۇE@B#*P((/W^=bgHD<9f/܇_eIGiX>89|3F,Vs@vq`m(Eiޟ·T[[_r}#&;ha(pGx^Lc)#%@Ҹx;hGώ2;OkO&&BADdE#!q_ygO9Շr{LJA5~]F2RzҒ21 :xÀ'_P2v>qa?WO~GElkv:<%_U]M)f!!6I$P6hbɲdHQ#&amFGq=#&W;[mb .qH|21Dչf&_a+#%5/glqgЪ}g_(Duy7bp#&29r瘜d2BY،InYQ))(d܉YA,U48lk#*&$hJeTN;;)S@UH%HYW>nC-o2`e;%'sc6{"㿔,C#* ԝT2!ϵ19Gg|z:J'>Wv4 l;5ټC%wEx{~WcjdoROdZMI?*?xl/-)%"Η440(QU(3((ߞgN:$i#E:]/'DI/+`aW0(]Ƨk=P1A6k5aCHvF5DT-O\(l@Ш/1M|.pjr5E)<uxM\G=PPp`}R#&ha'.v~q-SӘo]ӤC$&Թ^(X"$C !wIDF#*BXb(>6*HMͽhbǪ]m(+)G(Dr#D7sM#%5ۗ261;0odH6*/U!DJdh}hY((9'X!'^)nܼlZ@R%ꪫ)}WTjEdEB,eoyk^BչBlV.<- 27'?^W7]y >#*#*#&pE-zOy#%d7x03ìMJU ҁL!P(&#*pnoG!a"󣂖Cʨp񮼹pB;2lܝ6OI·f0A˪v2y}J6D+phnYTUk2qD6-VQ\%tL"|?mWTQAX*@|UR^|p)nL"By˹EJ!&5>BЫKL#&2a%B0r!#&QY%Y$[@^JC@KX$ʖmͻφ))E\P\9F@HeBVyGt/&xvn؏3C"4ulCE$!#&a?rbuw+.4`!]ChaHb2,"47lx_?WՈ ~VdW1`T>:dH>e*G쭢IөWCA=~Y.QXpQBeLk8z1*!p>$>wv\Ui!\;BͰ恒xBeIoip;#%vT6#%KSj:>TNa:s6!el@;W66\.S#%0:JEXw~EëPlzn#%UP(33l:!fm''HVd}%ZKƸu|ɄPlW5$M()jC&J1ݙ-Cbs]4Y^ywj)2~4W)U6vzm6w/ZS;[0;#&/_$Eqk6p%#*S$QĄqr2:+Pvz։XepJ	xԒL6kBJi:^ gCS@3;W0!2#&^6)aJCV	C!г>	ĪZEm{zo7''YjFv~ҚLQ:l9ۻz&Pv|hBRu0+d{`LMZd#%_Tyw?ЊPMqճfbxٽ"@t9#:ǲ6:<>=-CSy'*~*cMX,9M3O`h9lkYgX\h7CwS;uefݥG=cȸ`#*11.V]աA 659cS	J#&*}-ub%`/ "8#*͘Q|A	0YBpt+x=@Yz	A79gAPmm՘=#^x<eNlr0ʳ"fbOVId#&l[	uKRB0ߧz1_ ["#*#*H.PV7	f2L`,=!#*5y=NCg9.Rx2JdӒ@DsO8ՁVYKzAC4Q{*cvXF#*zI'=a^՟Z>PbdspPHbϒSpEGcBx\tT!l~iV#OxH"ǎ\fATAyMn/'aX pй0YjSsk͜}*Ֆ2((a\ThRF(n?H/IXw+ϟBSA#%]OGn?Iᐲ~O7~g#*Q&5@ZIrP{MS~By@wXba#&X@Fb`sh>$h0{Anw>9KU)C<T	ok5ݞAcd>I#&0WhD,-DǓ5E|)"q%vՇLhAmFư#&Q,ED {?˳5]j75T$#*Is{#%sކЂAwخԬoFI<rPT^\G_E>?w$4?#&-sS7eyL<Y,2'2dA`!In'$sC8]*jt QR&	C5!#*d[5xCm1D8p#*vbDVѳ}Z$MdD)^V#x0f#&҆\1K@gW=?VΠ$fWTj^J	ki:QrbP_{_/*&@2II#~*z#&&<8:u!ܨ4)-KIh)*kOյH%EqiW@K>fwa53Pd.kw_hQ챗#&7dp*!q96?PDB4sZ##*#&7ϲ1˴ưww{^#%I6C0@$PH5IV#*~qDn䟫k}(tenA M7"a/˫s٘(=E`p<5z0R<w#*2Hw83Sܐ667VIGIN<#gSoh1{/ȾnIYFxjt3`jL_G%<~Mpu;jxxPo"]͖iJvL豀'`f )YCsYp5sfAe\-kw(TnGg':D@<0OA_>这sOݞW5rO۫Lþ:6N}=(#%Ώ̫.L#c:- 抽"׻#*r#&KB/8kmFAbc`${2@2CP1nTLu>ۗ<bG6#&hk|h#%p&6N#&L7?Q}=>VSOqE(:_ktO~2:7n_garhj2\U!P$sWTe:om3S]=9OW<0{(=1/y7yaXԛ_{q3GEgޱyNڃVLҘfe<?cx7#l.#%Mhϑ Z:#D_9Vy)Ώgn<\]òGz六{ȚjǎtÀ!Msgm',l1lmf֥*OݎwjJbh4#hN:rhkdS#*}2#B̵L("ߊ\VkH78Ɂs#%#% {׹E"]Y+CcjUe>ja(sXV~k	P)	M47dBNT\tú'o[/ӛϥ5ճе֯Js֝~O9{DbiR;Rmb08Z1=\|Ud~E9ɂ	YH"2qz>:4vM5#&N 2[#*;C#%UʟGNnWR;@m`A_#*b.?vW$`jS@1 8h]*frp2O^ipdlҷ׾$)psw('3nz"71\MKtutaḹW"Zq꽳2c-麗	@#%orԛ|{NMQ(%=7iꙚ㇉^6ɶok%K˷b+r& #b7ƉIeQ3#*4O՝c!b*vQDGDܪEhm6C[Ht,>.Zl[-F	pi%8BPGEE.e,\OA>?#&i~=JԤ#*@wt k	"F<'0Uw?t7̜H_os LZ_u!?zpׁ^uUV=~Xֻ(#&:dNmNL_##%װx:;2HQvyDi$H?ӨScT0I7:4AC}:ۯ]9byC0;5hyo.*4وT]-v7r#oA	G`s8aF$48-zDUXTU*QM:#&wfPcw6#*?eu*d.X@hq8(yL*\:hO2w(z6+mUj)<1#&[VUL#&<v*MTgW8I84@ưB1х0`>#&(qn>°%^v&s<v~άU%@eg%QfL|s'yT#*6x߿V_q!NSq.{]-B0g#@5"r`:#%vx,SO;#*FJ(1(܅4"Pˌ(@g	HV{W;k{JVfE]fYH?u)_JCIZ -|0#&(>	!L×z[:b"<oig87d8-.E!36&9Y	{q[jg7}px!BaXƒTS "u 7ZiJR$tZOy#*0f# ?qd;gZZهW3Y|P#%fk%0c`;wm	5BZEYݕ<_^%#%⣪?_ȓ$GoV3#U&Fa*]}cc`KNj:Er3WvUc5]k<'Gy	#-<k[G#Ju:zz#o-YMǷ7Prλ.h9L5fBpmx5z;#*80սx<T㚚ǁJ# az$JL#%޲+@<[	F'"B4kuul#&D y|xI׾=nGq	aQKK#&M\Z}ahq4SnJry.pq	#&UQB b֦[L)kyzi˱9!ᯜ#%LvF7ZF(hw#*̱0#&#&D8{g/B>c~Xj[:\ȑ餘H@M&K	A,o0tlzg7sǻ0;9f> zGcG`l`gȣ9@0q(*W0wC{2؂뚀*n@3 nf&{#DTj#156g#*C11XQX%ls#M:/=7s)Ȑ#Ƹ^kL*nS-.HF#&v=$'Z`LI	0U4uJ#*H,5dm˹9,liP+q o+8NOEMR("c Ą5{t.p lךQ#*fbVfd˙cl̙̊J3/~9S=#*@7}Ys*"Pa#i\o:"X5OI+8cq0!zP?F[=[cfUhcEe*;w}!zzKK-E9Џ4άw3#&`&ədآA`L!#&L-FACPq ajiLdmUUb,qλUňjԍJ,!3:IlK#:H-6yS`U,iWpS54ۇaSE((lc4Bs'QQV#*`@}ne;ctls-H2*NT#%o݉t#*`L}]>%-uݺq	-=~w'@wIE惷Ƹߛ|&(g`6r6D!2zd׶7'BΓi[CEi4h#&:X˗!,>n|mOؤ&QD^g'	3'+U&؋m	%mn et0%PGٹơE275ίTD9L	m̛IeBYd$y#%hqUe2|r󥢩B8듌֦x3D%1#)ngMk2S5LwlaM&)M;iΧ{HIh@fz𫆠3p>qd~<|o<"m4I#&qÌ57#&E67#&ZKd\dV,vg9y&jfK#*p#*n~JVFˎ@ӂuS_nPT#%d"DLmMbMoJʐ9rAfg Ob{UĦTgN9n#С1X.2l#&FREp.F(9FTbyAb17J}M7 Ǯ6v䯏\EЋ҄=}4+k㗕hmIx8=k$}F0gopzs^m*&imq*,P~ΔnL?htPw')WQ՘@bMm38g|Ts'<&Í*c٪)rxao\\#DXη%dGuf3+cك|6iiy7ˎ5v-%gy껴Ӥ{^4EoR`bxP5(B=-QCPR4#%@&T#%9ݭ2=?!Yg%K*B+S3EJ3o[iG__p:RC_PCf_Y~)-KiT1,]P(A`N*s#RNSF$-(k6-Ϫ##%n#&uqb*o;!)g\2GܧJ@#ǻe*#%ft++{y\sBӤ:@d*W$B#%+ s<TUߴUL^B&jThaT=tmI5W|]ULgRI	քbudDa_XKpůiLR71#&FioLhcM"t8	"q*d_EWO) qAʧOdƧC)reu;o**n=HvY!#ssU$@$  Gb;nz#%AYAhF##%wp~And?\J=Pb6%>f%U؝j0h4#%Ln<)KC#*I4[v6.\+^]ܹ\ͼo7HGussIY]9$N%+F	u)S|[J}ɐhad=H{f(\3m<DEcr'Od+TS$N@Jehty`=1c-!8&@@d@(ub]ɝ4I_,#*j$|^"h}?οTň#%#%FA] Dd L`À{Kv#%'JGkVk+:`wW"k,X#*S#*Gs(ldX08BPm*ns0)%Y0\*1!HbnKJ?a~`S]0#o#&٬Gl~ߨ~0B~B#%uZ$׺=C]E{ݽpS\gU*~,g1F[``#djSQN;i~I$STIӳw,	ygKQDA!	fBȠ^|8TNRH-zG~F2{!Zʫ*E꿄tvX0#&ߟvQʩQU2M֔#&E>\.1Z-h2JIlV56LiT4Цfڥ""zI|g}#&|D$PH22"HGх 2fp6Wo᫋R60UG821`Lx(+,_*#*ŋ)ՊgW970$ADa"4mD 3&$]*WVJy@@.*ݭucH8v`nNsA#*SX{)(e;cZ@o˚ILksdZlY1L#QHWHqCh#"+H,#*:-{63amFiL.5^\$Cxuv) 烰vAI#˯_7qM$#%q!8ܠI!gw~#&	LUdcDE;1QL#**Q6a{;|C7&-uEĄeeXPuF2CKm3Wp2ë}u!N<#*d]{#&EhzqǿvjGztmBP#&͐Gݨj{ISIH͵N+XB%P05%&o|:_mdG^Ҁof`~7q5k-Ԙ	FoW~bnmϪQbY>ΧX(ip8VːxawȇSXTU{$$t#%PeT"g#%ځ'/@5>xRG`m0Ʌ5Y-sZÏLBHm2!#*Rz _ݚ1]SC$)Ȉ<#i-nchnb#&@d@RTT=#v'[Asu)#%#&| [|$W^TȎqUEBˈtJxR##*%TA(X%Ebx:t>\8)*lFڿRC H"0>SW#M4#&a`\=8PXA2b3U'SS_l:ҧe7~<R.J7em]uS t^)Y\!޵̏CDϟ3xa/oFcmaf6+-5ҍߢ`Gڻ.8+y/y1в}Fm6$ݷ{:q^Ω):o޳u]TdVz_Quʖs9D%q}7:דlAWAoqoƪ6##%Wiݖ6#ǏrHG9]LDy^[577QbXf7M((8yK6ӥiON4A lewC(B@%P"()2|+P:y}{u#%K8h'׫V^BEYl\bJFp̌ͦ#&a{k^:h|lIн~06pw#&WY{l 鎖uK@8\̅;4	˳߯}zwosתBlpnXw:PesH;eMJuEAc!rLbY)~7~VG%J>gSzQ@Z=\7dȂ]L0J#%zY&}Z#&t.2&8#*8}Dc.q۵BzFHghYTDdq:*{+\>,||2(rVTSGjF^=YFz$tnIui#%3cdF5:(@dSW~>GUt'Of-i@Ԝ@yʱ+^iMm5s6KWM'uu,4[Q)YAX,UA#&QD4^cBۮK"X$VMޫ|׷ӢԶ&BkLDhǀyp(9u&Vz}9ǧ>9ߧWܟw!DD<Ym%̚ͷu0PTHiA|o$ =t>W}'XJ(@^LRU$y,&jU֊F8(ж#TP%w}b~I_8M. W®yDF<߬'s`rSNMz*c$VUl4D9#"#q:*&#%-R5JmBG18Xn%(q9ȒX!߈|#%mŶjѿayp!1u_#*{١']e~Lm*#&_ D&ZZe\'3CHf[Ʊ#&-9s,X܈+#$Qh"My4/A#}Nx#ghbl72nhUX["ON}Ok2M!Gg[EWxO)(r6I4MX	:'3tr꩜{[ǘ 1=)kMJI6C}Buc#PfGYUv*vs#&Y%) K'ۃh#*#&f'`Pn򄏷%iMlgJ}v$k1ŝJ$0{Cùx1@4۪Z4KlHAAp8qHR0G0G(pDFrYЀHK)%A@@Ȝ_e+gƇ,4bA*H61#%/ARTEj#&.Ed:ĒA\ĻiRBQfP+#&0dL0PqgyDpHR_BAipPs#%ɈF9IQ]JRtq&ϳke+I\S7kXDDF=/(ۧ\굍YH=Ex w'#&\qOA2j3z%k%AusC<P$/aav_ڏ\ս,/]V,HH4ؕ!r0tpأK>tbY|N+Yw_`bmc*XNOmzfJD┸EcEh{*^#*c>IW͇Pj#%Dz]30#C`1 jF6ђ#1!>QlTSĖh ƶ)z-.d$.[C84;7Ng_yJI_t>),U΄:pfz}6aDOvN8CLKF#%LsCYRq{60;PuV#*̐YH#%4/ZڥTa!Sf#&8%fIvhlLW"JŘ/?ozCpy/I!腻mww`hޥ@rc3G*4f#CTXC齜0a05=ZnUT	Mbʽ	:P%V;S,m#*cg8(FSh9]9,1	rh3TvIyJk8c]q{Ñ2͉JmYc#&>k7O*T#*BA29G\N)<:3rQq5 `پx%#CLITJ`2*$D ƷpbǊ˄`"6|%yxʐ4fY6#&eValY!&6CWd`!Lهz6RC97#*ư^Y*%QPM0D)T2$(baV9vfWMC?äs|L%Mz}diW¨'01t&DcMBM	Ё,+S:˕KôbCCA7#&7)2qZ>TXcD3/TB! /Ue`5TN1CH [Y̺:]&Zu#&V̺XzeEKx:@#&Rqis,Utܣ eT	)ej2,}n;?qZ'7嶩[[VԘ*bSmx/sL	"#%)`El#s0B{*G a@lbF;jбC#&C8ihI}id`C0$#*j$M2IDF1$#%Щiٚ,#%!0S`V#*)P$#%DCxbɂ!WDZAY\rFgos;<q/$Y$AE'?^VydLvtr/<U\&Z+Nҧ2{haُfee`kN޸$Tl밄4O{O,*Z+o\mĖ*ּXsm[Tj[R@Z uF+93~7xZ#%R*'Wy2=dA))f2JdbeAWRfj1iJ)(RkLFܡZLhJfhS&&S2FdhEQLJeϺ66hR$ITfL,L"2`֘ز(X%*3iABF(UvvwR#	Lf"3O˞&d?%*))T/1UaiQӖ2:I-K#&mZe+$BI`l{͢~oAnmFwUr0YR	*ť?Z2	ϳ3C2~usAy5\6`p¶$|m~ECjf6VJŒRm2FK.f.^(3d{N9s#Xx1\#+[MTLgf2ׁo){$R0&i=I;WHxO9#*9Xesv-r?7.2ѹjK1>MG3J!	;,6R5BE@T vMHjoR/Ҟ4uD1no7yX9Ϗ?$bX[&z""/]򐠁0AB.Jkx u*iޜp]y#Xc~M:>ɇ<GېO5 L|f:k~xv|_lySoqDyo$4MX^#&u(jPʥ=OnuYBF Vn	օ^LgC|DB2 k*Tx^l	'aԁَ#%'M?{}^FHC߉:ʷSOUz{g!'_Z1kL9sH~O<:__/q*V^lJ:<ᗂf	Z#*ӾޙP7T()#&J>=	8_TX!vH*9`q13'2:݇vt0p8DC(#]2wC#>EFeWwgꁣ갬h#Ƶ##٘6Ԭ49C9FG^Iަf~oUϮ7mq9cw]2W79aM54ʰik5J45cx[̝rLML4{7wRhhr'&ڡfF">0&f3T~n[Ȗ,hfSYdi6m5KgR7e;IY,Y5:88q~^хƌcWmFg*3#%gp27CS[K.bl,ཌq0@@eh|bwvvhloC >[MYY5V[)txk͝wBufBPDũnl:\YӨ#O1vufJ3:zXY?9C1$Wɽ2eŃGZ-]dƺg<ƍMd~c}Aj!v79ӗF"LsuQVgLСaUnowXqr	ݶ!nAYp'OW)-cNDs}%ׁ$Lpul6\#xC[aQή&GmNa#&6s}mb4Pb+xzݞ#&ƙՌ#dǅ+F(i>X qMds9Y! 08Q^,҉F$K64P`QiZ4ݏX8zs#&"g,SKV5EDU{&-6#&Lߣa	n<)xԣCCv)2ͺV:&Oi7e1BLXE2\ͬ0ewB9G)\⟃Y%䦼69ef!YgghDhI.Gl4Ӿe:.n 0MZ*pq;>cEw9#z894Ӥ2u\ig4HlipXlhyz^-IPT]L#o#&ޮ0u5ژMWsth4urr#*mwL-zтP:jtq5b`WNk'	x9SNA&JmW׍pԢ[n\,5X|'5mcYp@ѦV	ĜmifİMۏ9֫^D`<"jYV^HhTTp$71ly1sqgqIg.`ި5ej);>"NW42^;{*M|N5&tz)8dGc$cmI6V6gZkeWy]55,y*fCIS5Z뽲fFm&+,7`>WW6|*#&4	s8VSpϓ{#&K.Vj0gf{`Y2O(N0{LQoDV3H皗1Sd"WbML`ihm]NpTeLZ.cKV:)1+\BGT(Chmg4kI&Wp$X*:Żk$Jv(jҋdKYA&	AJ<<p#&VN#&%BNU#*#&o6/%#&ۃe":ѱhmEcu.8w#&	$aیTR(#*Lb4Ӗ`^"XL`F%.<`p e`ڂa"qB)6X(ce00elM$FL	lW%Q](k'3sKަ*ReDI$=o&A#&LGNJSei^3aTc8z̹LfBYL("<>2@?(x}sJMηieX|[!͎r-}#&Ԑ6ChM6:`Vi@7ڍ[^w.<+4beLi03gq@X{2a!QD=]kcrS]g`)" 9e}܍խ#*cRGsEr׋oCjzmsy$#&!!#&G]¶4T4mIÝgZ.s#&&qW좐xPTXȎ,CoC`6֢-1/nho"JTFPqCT6p@!r;#&;"ɐ#&F+14b;q/KhJU)Q#%lWB8&ǙP#&u6)FJh	r@DqL#%ͦJ3שPT#&!&a(.ܩ#&2"a!t%c@sdh*I5@ypP[hl"阦SilebX{Kwؐ1t8#&&b]g8Pnl#%LMPYJa1 J?aK!CT3i-5hRE=IJ#%Ag{AC@W$%Xږb{ڽ:#*:@	3l+@#*l(ELfI2evI%H;u\VO-ؽREQfXltA\4W겑uBX$iv8:n+bG$5Jчl,b `8ٹg.HxnX!ݍ#%{Hc4#*	s(R$F;@Jkmݺu,k$i7DD"X(l#C`@/Ha0AZf3B%`w#&Jbh#>8D_ ?& 5//]̹qcmRS"#&8'P'aG'#SQ7 P70A#Eê'-Z:p"m#zU	4.s̅Klq~5ay`#XL&9g6H44DN6'ɥ%/#nEw_fX8Rh5S NlxkHohjΑs.7S#%F;@IE %h?9#%>|BKV	pjp=ʏ$(mj5&iRT6a+QQDYfv	n{vt\SwGC*݆֐$mEM X	&#%ZMPƫA@LX#%-׻@yQi;kְ.@Pxؘ)2~WEm)$Ód+00 j5\:QAXbZdQ!7Hыc6M0U&1lR#& 1#%MS(@Jk CjTubi6dI#%(eչ3+sGiC<̂ov~Ѩ-/sO~.]A!PrA{M$1D"wAk`Mi!K-cT ^)"WTK&"jy|';nQrVBR).Z~趲 TVY\#%`3KqCԢ UUFjO"l I&pI#*0kL p#&􋑅))#*f/W~QB̍G{դ)6Jk#&hi]!2i飖ڻXE`"`;`vhT<4BAi͖bSFSCӢތwDansVZb8N;5؇dla80"g#&/2SlD@r&VVCStSJY5skՑ5:mu9[hB<Ѡp24BjH``72&HȘ$rP[p"#%#*\	0uLM0kCZpR}d/fLUi"@?)wB"Ud:,P5'dw`VzSYy"dFHյ⿅Umdjh@!z^rCsⰱ.GO*PK;Y,,կo5:ѽe1*ĮE#&{:ii4̘[MLRYk2ι3T!ԵgX"M#&8)hFt3n=8EQ-<8q^jdZ]-rF1M&QY\5&0.] gC}yrζ[qpіzW8ߙm#&Toq^Dż[C±J|!t3'\IW֪gODbL$&4e)}Ɏ>1<+.f<1ȺqA66L!R:6M(kjYlhDbrs1|tB,JŌDqX0cfEZJCRZL	h`E,@ ȭIt)Bb},Yؤë#&0ʠc%[o=.BX6uNu˩q]q:2Қlj&nWuۻJ-y(izwkHV-[Ti*Yx88)E60I625BRdMjjE3]Mɭ)jOj*̠hٛjЕYH#	.j<hL:1	&!d!hMͣ.Ģ%"( m"X=9a=$#ǦtfgՐq;Avj	N/ل&8xA,;3#*4e<-%#*4{۽LG370ϲ[o@16cIObCA0A1XAIÁgޝȑLPN|:R<.FJ*l8lZZJL@DP3$4I8v0#5e>F#*=Zg&-YPE#%%.8<EP+i8H6"IvgsCd%ڰg>	!Kʣm@Kakn	Z}UNrТj;.=!? R*V|=?A#*{Ox'uYG#%"=b_Dв#%́Dd"(3D=v$o5LVuZiJ[IjDM2ɦ)XSMh~Uddi-mVQ'[vn݊*b0LmLKSMTUjhMUr5I#*[V$jRѣ5I55c;4iʕaemho:%*(W+5IYoꌘZDlK՚qx_JٺXr|T\2MfНq;dCnq}_gUZ#TВW)Aˇ,F^ַ*ZZYAI"K낸d"5foem[XMҘHY#*PDY#% #*VQl$J6AհDHjB!QŒ-̖Һ0j&iS-d)&%CHͱ-(ҔB#&&S,lFel6Lhllde+EJl(dڌ!$mTjͦ!BL`M2M2QK5SlcTIVDRɍ&ɩKljjݪҳkMi,KZl٭R֒TZڼ٪-Ͷ[FKc{*kjDSmZ6hI^.'/	6<^a`=ھ#%()i8{UDDnwI*O#5!%$!~dM?^64tMo4$T4K#xfλnҡedgo	eKXy 'p}t_bȲ(!/Bfޗ9dy5A#%|RsA|#&z`mXlaP;{Q62rV'H*`@L&JQTpXCv$<	T7bca9`vr#&xynmCspn&dM#&ߡv#*0*EKD~L~l+$=TTͨUrCF~UG6к|k|#&re}gB6a0Ca"D5Ln"H<#%cmU-]URL%#*CTzB*ӕ+W+Պ#%{'lͧ"0cĥ+e%SA;ݕN65s7#%8M/wj7:yWNl`uE >6sp$虪X!B#*E>=K#J󢮚+?iT`8(.$d`#XCm^&j#QOw^*7H^ŨqVՎB$Ürmg6&Fdq)"I<lb0<,0?=!)V&F#&t#%37[trǀHбv#%8uPfE9{FRR`S`C6gܛ7۹x)tכI+2fkIÓTu֩~{-@-2"_Vht/@ѝH|Ŝ`bݳw7t~a_äSi@0a/W0!\#&q痜ڍ6?tBb #&k01m\bT5n	rB,ycf#ة#*g3YhVI2t-@(Όrt!NNB3$!{uS#UIL7E0c$NZ!bIjHlj먊hnQ9Fȅo:Zp{R\	FKPaA\ֱ,Ӄ&y稌,⇚$,(?#**BvB6ks}+CNpAUNâE$oŨwew}O/9Ϩ&#%i4 b/ʺ*4M:LQZ*h*;z#9fGU'>hXNt)EHdEr{xi/;[ŧ9f4j"S&c}45hφ^6#&;}Z&aqcdǀdL4u#&{!l7-"DT5V+颕x=E,3VT.4OFN!ÙTY#Dc"5uM-Hʸjż?GnEhTP#%H$(Q8#%Xcm|;[]Fg.=8aɼbkM#NGtp,rB]܂_:	W健S٘ 44V*3wi	#*8WEJRKA~zIB0yfڸ71(@iIW/4$l<8]Gvz2Z̊]/[WG!ϲ:0M hT{5|^7;m"+ı77e%T	>7I;齧f2LhHK#%bN(".9P0C\N%J`ǩ#%6j.5$2LrD1Cp8"B6,mjzKۼNd(ĸ,,db5龵pH#%=e-M	{8g#%PbXU;zHY+M_wؾ}Ls&߄F"/5aW֝Ցp2Rdfd$Q[٦+TmUr3lIC;T[cm`T[zȢZ5 OAEDa(Q%.Ba;_慙O͍,&]v8#%F_sSO;.fJƓEcyΠRɠ#aKQ]{4Ѩh,#`6풊jwh-)CJ)$#&?5RUOMTTSe<#&y9(?Oە]lZdj4ȬJPZ#%M#&ݲ3#%k|H1֦k(`yw#%CC0%=a6l&JBZJH̱caO>&Yxaɀ.=#%0" 9{>&0(#%#*Q*nߥox2pE!' \߯H 5_Sy'Өc[7#$%DrA-Ź}	lXr:G8秼+EZ}-~~/)@TT:CBXD-=#*;e۞r߳8LHB? &3Zn5E6 c-:dQFEd;E)QVE)5THFŘTDd,[Hp#&|/жݿچ#&-u(2`QJ?xϮǡS}8S})$jzؔjzqNXH)$"$#D8Z)ȼgHeTU;}yRfJcwf &Y`zFQa}c0͛.PJi#%!z!Rvd#*]R3 ϴj}O*+ޤQ5kSt**T4T2HIK#%Ic՘d\{#a-+|V ā.!6c:Kb7,R^#*%&)mY](PZ@EP3LlPf `@DŴItʧ.eY[h|V5x@4oZC"f3$4d%2JT##%G5#&3R*,EP(MXXD\5IikFsYˌ#&*°&P޲,	2ݽxo$F7ع6sQX1$X^k"ࢡ5QRPPҦs Z7Pd]ͬLq.#*k*:7LDbg#bnE" 8PRrINlvIL$[Ls҃.FA@CKmryT f&Qc!Ϯ4٤Y+1k%#&4ĺ1>7Pd9[Og/Y9"oDGTA=ACXR5u#5?$+dٌ6ݶJ*%5mԬj-7kdҨH##*@M@F*$X074#*#*QxB Ρx&W	?#% G՟>1₄qC3P aZ*F"d]X!aوH믳%9lMAAօ4I1UI!>T&MfۚůZ^&mfmұ]Wv^[QmmyLL)&&Ijᕯs```/VQdyEX9uC"`QI:`ƪ!u#%@0L&Y	e8I%﨨8ΨO{#&M9Hd$D4.<umB/1e`*BUgB=t(#&v]\=غAC#%#%k>178#&\z#%%Pëh;BѰu6s)(JTƶPM{v5ҲuA%`[swMǻc#%0a6ۛj@pߥ^zF;v9ͥ4GRC&$MiR?xn^c9WCkU2'M0F	Eh r4ʑ8Q0cZ$f#hxy`,$4Ll`z-o=3/IKa.@0TJ%D0_#&$z-{Σg4~8PfIifttIQIԧK6a#5q$v#PJ>Q\3b9$b JD3JQ9#%Y~>#&/X(d#-lۈš}~h<:zvjCLx5O?&Ujq0<X`c2l֗ꆌ:OP zrjyVT><Sz|Q-1E#&RRRi-֚֌fMT~msU3*1j1Mfık5lTSVMKkEje,֦*F6>77M0pۀk4Lp1U:뤫rjrQTZmskx6#*H?%%ջ!#&[۔c%ԾMI#%$WӲ#&A4RChe>J>x'd#=rE/%@;>!wd#%QmԜ#%,G#ֿa!BD_(}.D@"(_S<Brc(wvi!#`#&|{GfAl)m~\-pK!6,mzOA#&y2EGhovU6fYwr"zt܈06Xc:5nIK&b{#&]Z`<JfUQbw;,E(%?,qβ#*UeM|jИb[qN%z`]!#@<nO:=#ZnI5`U4.\c"&)z.g>,bl4"q1~#*^>&xbbNOf4j#%?7wVFť6P/umdq#&0#o>0EL.aU-#*Z!#&,zj4X{jhil*@@.Q[@qSo#%֣J#%,!tlDX0#&59d"Q:6.7ļX-}#W:ayhmre&#&GásM<B\w>ۛ)n@^ͅцqDD0BB%BȔ8""xQ=]Ԩ:?/N@}THȲ8!MkVIO2zFfF&!Xnqh^]*3<Hy=faB"Y*7m3sM|}Onw=(H#%@J,{З&I!wyM䒞^6՞>#&@PdPĈ9T)iPPm7.2fi h9ZN'w&9eH0(Dn|o6)ٴ[C=~'sLevw~ʯ*>zE`UV8hmJr"-޴50H$HGIMk+R~N)Q<'ٝl`NBO{ez$6~lnQYfmИ?y3z$qB\d9N*Or5yy3yqrt|loIK:vQ6+Sq|7ۑ>"5_V3g3M3r^a9*Μ]a}(\4.urir-[Р*fqZ	E(`%P.Z$ 0ʷôL+b*jDuΟ`"Mʹ>Ө^#*c۫fNSM=EÄI#%YA$:)ׯ#*	*bX5U(j~\5͇L&ycB"_o¬@ ֫^oF%VAZ#&W"v!12:0n4THB(,}<g"`b?H%~#&X!$h1 g#*1LV%kM%V j4>gku6޺;i} *eLk˝͓_XYVMF'S2GH.jbh5b͇2i%S#kPRF̜8kj#Can,XAwRPx1[+7oC0t ;6u<|RG.5T3a,XICLTįێ6%S	:*jU^Qji6QJٖUSU+o֙]v;rەA֝wc7,6mSv >Y!PMAie.E>Tthr0.J0|T9JM7IN_E^otXjO3[=d3	L/wut Q[$	T@MDa4q|&y3=&H$Tc,ݴ?Ac<"[V,S7#&8ZaV3Joy[E2#4djϙ;qm#s{x%2tƶBaέs|I-KW,>i!A<A`Q.:=}m%'YF#&)M09l4l{b18gIDflSQutpYȚ:?Bs90_(.%mwܖ6NP	/%PHfyaB`(/#%:!Lu3F#[/cgYHZq@,3UUT#F#%6)h p٪f/q~B_+m#*&Glh `#&ugުVڙZޕ2IB#$5+qpriVgM@rMZ[:_?S}Vd\0@P:J*̡P #\!iul$=4t..lECt#%x"8tF:eW`H,0FfDz	*$1 Q\MV#%#&ur[[M5h))fmjcErx&ڭIC4" sHuNhx1]S-#9~P-lzY8cCTD^m;t8;ehT)#*1ޡBFD6aѫːjHYuEՖv!yBI#&nM#*16f߰6Գڦ,Ufzmu4A#*y!c/V_:`Mj/y@+u_>U]*`(Raf	Nښ&4r#%3$X8	CL}_iJ#*T_{2N+q/6A.wbK$DAؘ^2!#&eht:hKꚦR䍷R#*3P%0Qp<`k#&(an~#&Жݐ+#&;;Բ>8H5lr܊:<e$;׈K޳I$6ٖؐ2}7N!-o	2Pi|D/xa:a.xPmW0!f@M`#;þ^-:c(%c-Gc,ۼ2|Vj1q#&ߜ^@Dl4:#%	Ip`cAIl`#(O[zбq_v4İ92s>:{#&fɏu#XT96A`8ǻcd;@>Ѭ:Lg=#eN~:-X@az T0xp`xnU]Z]0CUZkmQj*mkRmckUA#%RLav:WW+F#؊F͡hL^pO.#%@Ag`;* C*$4Vz5pC1O_uWC2mFmCEli @Z'ُ݁:(g5#%#*hzn#%mDwK9B~u<|O#%tj)!k(Nd8g}Db14ٮѠs2G/Li1VAdRПrjnҋ˽%	u׽=&@f]#%\#*pc${r/@F	JI$;̖@1 ҢGS#d",AZ#*5+!#sGr#&ۀa1#$0MB9ﵔ:r5iA8D$Lr!5HEOMzq[-=QS$;;W,.P$$e~\T684KMY8uib!|k\p,ESUO<EJOBo^Pk#%L8U@83mR-)QL^kF`[#*>ѨH8%FHْSxy7^[ufMLͣacQ_X`c`61fB6048*-j)b7vF,H4ѭSH EA+R#&!1\؅3Ģn#hlҮo6MZJKݲD*J`d3F,J	K(@9-1*2VI+TS)V%iL$6fIRݽ]nM;͹*)&VY[0zzvl2ieU1CSW#&Vn#&&1VlU#DPhǉW11VCnHkNTEZ8L4TW$YWlh{$q=[{ҁa2m&R1iK0K*H,V-ծE+5&]'X6fY-V-]#dbRBZ#*D1Z̑(k@JJlVR+UcxیicPIiZmؓvA1vn\Ak5nBlWCҭ&+Y!/Ǧ}&hZCҙ~LV1GzgkIc1,[e©XS2jv6&̶u#&%&"ᅬZfPbT^X|r$}QifVAI{RyzЇfbCBӘesU+Юm`cy9bzYY3Y261f%oYΙޡ4(væ!E!A#-n=(*Ր#%q	vii=D[̪lc~B'l{,@!DTZQ!A#&-T6ҤT0#%`1A)#*"d"b&	rKEp;7ozt55,ͳ!n>sX-Xp!SG鑢boi&kn_KT&JT$$E6"4ZB*߯=bpnli)dn2|#*:Uj̔^]	d뮧:gg\*#/66o:+x6&؂}#&n5kK}<MEsnV^۶"~zE	72ƙ0G'G1QOsνqB`vjK<17eES-#%rH$64{r?~8ew>i/RنS)x{=ߋtpOpd=*( | "jl*6\j<RI!4;Utpbox쭸[nkdcWr髫nUݹ˗SI-#b6#*h%eSzm^2mi7Mk<bэ[dpȊٕ{Tcu2Y]Z(H"@V	kDB`#*$ˑ6!@QB%F$L[ \Qv=-ee SRAC,(Xռjf')جe)cQhT[EbcL64IʦS2Xƪm]e)#*s>$CH}_lDQcU^i-&c*>]׽2F-ÝaQ#&%4}*`hvj#%(HB-Eb[Vmtqy;tݤm&[JVֲuWTQMCk?Q@>*I#%L#&2I1R:.I$AA#R{W56d\:SŢm	=#*uڞ}]'~ʊ#&%]V&DX<r>ӼGwuQ͸csI 14&1;L6$id8sSonR7i\׼霌cM&4#&toqwXBC>ҺK*%	L|TV|6we$#%9բ.0ATC#&1.gUX,>dXtܬnc"6N6[?N6ƇV=*&#Id$7C@adŵ%ʗծ%`c?I(!4@V,G#&$Kl0;] wQ?}#%Ā˘#*CΏؙ8#*(aR$DTKPw4@kРь#&ՠꘘc`EWWYcZfؾ>Uuv']H\מ];׬%X$c!d8hkefQLS5jRid-XќHa0(0P*-lX#&/0%f!L"\^P~{F)#&1#*"0y?^sxv%|0c4jg4{hq\22qen8|cCZdɕl(s39`<#*[_lڭ_0Jt੄o~X9ECC8p>e6w6uF|#*3g04ڊa誰uӹKN`#J&}";l<?b9.6NMwn3.lx$boH-*`b>)E((KC,Q<lD	;Hn-{돚~Fwdi,֨kėȡ4K/EU)TL uE\ӌ	--eUn(sa6ٽZ0>$T#8I˚̓z{#%?`)UM*HHsJU;(BT#%L2|.;#&szwUZ2?M\'CcHBҪꈋ\/Q2qdMBiMQmڛdT<nsj3to.ul:+H^-Tr[ifX;y[QVJmXynRTu%ӛwftܑJcu:͍G6صHAj2"]ROT5@Ec$T~* ?D"H%	K#*EubNPl,GW~_	fz2#*{fáBȣ5fi|m~"P]irAQ3_y[ZRhɫV-+#&=ŘP$6%U:뻩eb&F#&#*k݊ƘC܄DM2D	jW"#tɒ(6Km_Gr\yWzQ6DmV$QJ(ZR,`ETDd"P`7<n˘%	TJΙ?ѼbP``{sQbP;QE]b)`ŀ@	 ΀a#%ӗ#*iX B,@	"H/nC@~QRj-2U5H\apqijgM"C%9AM.Pj$;5	6aszW߳o]?q#⁆</0xa#*b#&#*T$.PDVٌB;0_CM'SДv|15$Kd*	Ǆ"@@XqAEˌ$d)B"`c	opPQ$cVmcCct%#*#*U[Zmm:*y1̞E/}5ЪS~A!#&RH(ipux#*MQhhTԋ)L*ȅe1D#%R,XT#*)@-A6lhŃ#&MvA?-?v<k>a(ݸ0*~И0"PpU&"sG[#&2k@)$ 	y+~?nleKm.]s9?g fH1i+\^eq6ČQi	?TBGpdéA~E?(엺gjKV6(&#&2[\E͌d1UmS\DxY1BHU!+?ơihQVv*-5#*mWLWb^Vx#3GB#*hD<o#&ghyPOO:Ѹ~ [D##Y#%$) CLUE!!Q(|]d{*SL:.)Cp;*#$0$IDDFkc*(!@#%Q~o?Y(J,8/rt<)5 COw0a-e4J((FBvt(ԽLKgdd6;*ɤ͋1C↍lP~<h(΀rbEPWm*Tp,fщEwx+ݵD?#&C0r!Hx,Gk4񲾘3rK+yǺajaUNls$P4$~_"Nӱ,[Ts	_܄rSRib)wWqν2KDER ,12`YF6MSZ>LIyU&]N<#*IX`LwF}/-żޘ1=C5 'n} Ѵ#&3xTII*}U^=0@&`Me!đ7}fu^ϤAV>ŚuTWCwjqS\"hx?*	poLlGSJa]l˓GC&#*	""HUP/MY,WD.Y-ĵ;GԝHkugIAYTm<M߱&D#*7h+tHZj`ELạhpv#%fV]_fr}l<]XP'3_TjkADi~gZoAE-\}ˋfj)c^?l(WȬ颗',+{^fǏTxtrƀd^_+ƌ)	g!`phDJ"uh#*$2׉շZgj\ JnN-W*͈>{JfρZ6I$#&LJ	s,]`|4BSbz	]c}I)U@E|j"\PD#%'0O#*O#%`ĂH)h$Z<|@h#%BDE"ٰ]XuyIΆH#&7F<C}X};%_ma-<&&&Diπlzϴ-`v&*Y{x՘JZ3;!ɓF,$":9C N"XxBEW~W\L8GÊʏ]WpWr8Y|ǯCX?#*/Q$HB;b,R1MuiY;rHo#*(+hIKB1)	WYc)H	J腣i@18f#%Rj9#%5#%wFu.zH8l-_~fC8}LPGfh_<QDS̼tZKC}q#*A]S6Z{6%e{n}ԔrFLi7#*4d@Zr%B	s0"0u?ʹ@ܚnafG{^^UReXcrG2%%1ܞc.\i	N|C|=(M~δIfp̻$;9-U&tP8lQGZ vxiØl8Ɛ6t?^.H8;ڰѧNNkjXe4lر-j"ud6iBP丬L;r><}MXmUo^ERl$EL(%-i(LQ( қBfɲԻH	A/	YL_\wjRInٝ6#&%d82;ɼnl &]rͪ|*Λ=̽f$/Q]a(!<QʢD;eJ}rѴ wiD+#&#&[Ccrmvig%7^A;G,h,@e1մ"4qեT?. !&̝ҹ֮8%A4܀DHZ`ciF@eb`IwHA2Urҕ7W*Mݔm#&4d%-STEaL (L!X;A#%PBT\"ux#$HH\#&tgujz7#%63*3)$bᤢN ʳ-vc\TlL1ZUC;. fY%J*9sGI~˷PvjO]:#&#*葺,HJ!BR6drf2ydB۹ݼk}"$tڭH-0i#!HVS*iTYFZp1ڊ3p(4ՂcܘB2\cxc	"m+#*\'`#*dcm#&6RԷLӣqDFR,܏7x̻&*443SL[tZ5<hGMF5pŴaRV+-Xj.#&>a۟8h0ATIAfX22!CMk#F5#&1fqG#*C	FLF#*Gp`w81`KVQT[sD 2-A#*ÜM\!Je~ Xrıv6`QgVtbRW -#&,00spY ,0S#*@cV]^mvTۥk(ĻO옃{cxfE_)jH!D%B&ת>-1@CN	iExgOt̓#&*5CM*jl[n2{GWW0PAb{iaBUYd5)tF*'G\zt&[YDp*.`Mn#*D#%4HؚgS#*hF6(@Ki#&ĕFEI#%i2@Ɛ#&#&)R#&*It:$j)(mB5ɒ)!"j#&vܹwⴔj-(Z6Țn?(;A8An'}|*#%APۜ$x _y/˨(CnNxIUImY[ꢅi"{*RT^Av̱YE	}c[m1AIzrqlmUq	UB0C5J3s:yJO]y5|o!Āu@pf@6q{׺,#&"#&uyӲ޳ۻ'dYI㲹AIY#A #&tbŦ=Nqmq0O1PE/*	2Dߎ\#*MQ(H.dU&4q=mGӈ3\@:>?]-:~d~u]ٜ{60$a#&/ǎ%POy@!#*<8$ w})&.Ye8J'N9XuTӴ#%J&ա#%'WX%S"HNlvXFV4T~koUxxʪ0>Owkqp꬐fZP$ԁ:	C#&)Rks#&%ܯ8%Iv-"I'\aL%7 cZOKjŮoM[5msW+<nmu[XVCXky6ƲkFvhMp#*`L S	C%BARkOx]Me7#/dhƔ,AQsTPZ#Ĉi#wf͵\?.Aj)RG/d@3#&>	3#*E#%>kq뻅ٴ⪄.H ))`15*EIdADQiP%ȪAp$$#*Bb_.[ilڥVHUAـ@?#%'4\:ȣT+#%|x (bcyRiAx8$#%#*f/f,<5`9bWʤNT<V  XA0-,-4biZLm,2fJV2h-X-%b>% 03 VȂ191ȑ>XR41yC)GDDAP#4f(""pȱW#*bMeldEd(YR&S1TB"@ #%!V|AB@#%OL;iKEKśԻfj~$LAXƅM@Y/n*mg$SϩџMisE)^>!#%Q ɔMpx#*aLv|,,?;|GEz^ah	 qҥćhO<#&MaDό:PQ"K+@d`4LgɈD$dRA	lD:{&BC5d	JX3b_LV4@#*J=ZHw,nu9xqhQn5LW*r΋Ձ)7N:20Px}Oa,֝xWR?m잫#:\n5n	z[N'iu]ȾB;R_C)FEiq8i6s>N:#*ֱ<vu׻[?7LFi="-̣߭a}6V)Z*jZ~#%083X̪Ԯa^Jz[&v~:c!{-&IN7Gm={d	ߣ=};i"S$C̄g{i{ۜw{n%	1RtO!/9wqJIhb~^_?:.(6R	)j*T7SH"K7@^jQO#*\kÒNdbaq0ZKR=?8vLn8_#&ޕaG?_R`.?eJ7SpOeS@RIuVBYǈx[Mb7	\&P)ñm}ys9!{k)a#&d87PC#&F˶o.k:ThgmmtIDm5o^H;&M#*:Ή]S6`cYE*,3nUQy!b|oj,:X}2ĸ#jf)nnoLQf0F5vd#*NYm 8~blv#{f>p׊E͠m<po~lKhO2I<fj7(m<)بDtvRfAZ٪v#*xvs>U}dG&OqN=bݶ]n_'K^o,s9^!t)戈{z]߬b<ٚN71^Tmqs\mOL홭<zAzDh(JyNaҗ[lu;t#&c*a}J<y/7ҿ?WPc7^߸m#*#Q;qcPm֜VpnjZ<03h mE/I\P`ɍނlF<M(֟Hd<7cz9ZeO#rO.|)zpm# xES!dͬ1%qw"㧖H7 ę*&HmDm]D҅4,# f<(ݠl]`^E,T9$aJ7{4@pl@Pfi9b|t2=<PxJHmLhv(pbP%#%Ra"<uuoxvJfg2g~	BUy[dC]#&Bdi#&g獌j9e7#&oPuG?${sǁ6Nvq銦NdL{S!"~#v[8̩Qt}>Q[u~ʏsW2Һx8&[K}tãw໣ek=1b0u '1gooi^39G*t!*b;}8peBuNTB<FPpKRmJ01D qĥ+-?tSP`t{U;nvsp8ҵhRJuqd:hM]qN<(n͒eL%m,:-_m8/5V\E1o\{Nd"u?);T#*1fL@U>nKJ9dcȇhll/`b6ʩj;{809b%K/cKV3R`g>KTʂ*'(*ih}੗JQd 2p591xX*1z˺<wds;`fKGBtO^#ܙ#*GeD4#*NzQ'j,$)Ŝͥ.t<s3#&;g)\C<,,F3<"cڶU#Uu#*yaϪi5#&)]R5*@"q`\uR@tKY0n&ytCDH#%nPXb\GwC#&궫duYء4xy nVf3"-BY`l{XoΜyiuS^pǺ0"v'_W)4pxsC-rZ0$gԚ"y#&cϲ^M{";Ɲ؊b&J&3h{sy\*#J׋Mrp껶w[uy^6ޕPlkwj4jzj-[WHj[kxܵFJEl#&)APdTDE#*6`1D-X@@SLmZNA梊'^mȹ,m_֋kEW5*^ܞSwRʷkǎ[--V1l1EJ2ew0:[,sAYӒEy(ƚeUrM،aU&xum<	b!!H4WAiDOR;]%yRRI\fMFiRE,%ʫƈa+y`YyG;o[5+hB웪#*@T4pQevՕQ@(c^z8cZh;Yf#54QHj*=5ks).^N1qVdbCvU^,m/"*g2\$cI!9N+:m&{|{y]Stn9bǻ4FaPHȅiZp$9ʦs+VLKlx˩E\17 7@+\ФI&䰍ݵvofmb5ǵ6R'!xB1w5o4M#C+6#&pUJakq)``VnK,޼YN#*%ef"#&lhclBd)33k_L%|4m\jC#~#NH$+gH޶a՘66#&Vm.!l)BWXc4--1N9MRV;8m,\`s8dڔ<+*B08Q#*|*fcʓ"E#*=HE]A0A΃S~g5Z#&malSL,g	aZe24pJf2:ˆM0F(+3+c(MCč%{rgo]l!W̘E_2q>S)GaeL/`iey@)yX);!DRkC$}M"c/&=%88"5mAOg}+iYXLC+Hx?p"6A#%hnGW2|#%ǣ	ºykf"WsPhFOL~7-T.S3osuz=]wIܘD`UNygl{νIR#9*{h OY]9Fy@K_RTjkWM[k˫Ayo꘧"#:绺i7ĐTIm(,#%ASI#&TLdkmJ56c6TXMiP4J)J!%DP5R&4S4l#&*Z@0لHa$b5zw.wp=1(=ܧc8>]Bk/F		Cy#%M[_o$˫h'(,Y$C*h" 6R;{#&CF.5PMMs:0w$B+	n7糠f"bK|$80A w:A':?J]V)k-c{e> E-5I}kVa&Ճ>s6XHY>!aP#T$(y-eǿplmB;E4n~L5XGCEh#PEr.edA.d-df[*יF2{&؆kQߪ,AGa;^e*11&FEhJu(׎8G΃"\	Ԑ*Ɇ#(0B	e덣oM&bW-q&2r͈^/yR:40wђiՄ;\$;	!Ye#*sHP! h_"0UV`#%!ɀ#%zr&#%a31ҕH#%Bwǿ8ԕPY𽓬I˅oى1JA:V#&?%`˚uL2rrUMAՄ.`*wCdzhVF,6l'¨%h_o4x4I9#_qPu߄i1#*]V#6O'$ϻG|g CT&q&stRA@AbۇҼi1hEeTNꋭ#&u'k^G#*X͏i!"0E9Xtnw[dM5g5.a@F9^0l8`,vqB	I=<N&;*׈q!E(Zn*QəѸzc0 H+AT#нJ;}׭؃*>{`aaLitGC:Բ]4X`C7eL+TLG"=;j[#% ]7#*/gL#&n{-=t:s$o(S..)k;^X^)L>J2T5>ڭolp<.ap"R'ۦgw목ƆP;b3"`PHnаdhi)D>xg*)y4s Òs Npc#&bJ(RKM6+BffczCj[aK7,i,0e5.c}elK+d<[J#&h$ܩ~%$bjAclA+zG83hJ8ތb]YS#db#*8"2""W#&كz]*Tf!6L0lk3ZHf /P`5Qt~#*"f꥘59"1γ^3i}Ȃ!8^klWNC<sxqnV	$ݕ>J۸t䡉gE8.Ӥvk)Kg#0o<dlۘyBf-"ɣ%*J+Hł11Hk6=D#%B*\tyEI6>QJUŅ&IB$B#yi|ʻpHX-27#];7Ykϕ!?(Ye>xE.e3BMpԀtaUFa|pMkuDuZr*v]ztE2OfJK/ٓ':mfƴ'9%VWl	O\8`et}gVTmVbf0(veu뮋#*5QQ	JM-}bgcNJh[^TF#*GQ:nѼ0ɑԖɖ44B0<+Z&KNӪ/'xWvmNRFC#%bX:&X#@G@ S_nMqĥz7j	;)q=z'N#&tLS(nH&lx5-陉Ru8U7Km[?$0[;.#%XtAC0E1đT!Q#.pcKspWXXKRq|udx#&sZV1mqZ`w[cNi٨}]N&DtRPh!-!DZ:ntQ6$J!\6E4X1Wl݆X""-Ô^@ʽPbmÌ{u&=5deJ_B3RNpZp >^B2IA1):IERZ6)f娘@m8,(l1(Itpw];uV#\kqƌuY1&i$8#*hos[55ys_%tYeTsQbY)N>Ԧ<pW0@@I#*1#&) ~3@LxMkNɍ7C4;0#gl}dsc$)n;LВAӁBu`Tkjz#rB"#2G71!0̓6~q[,C$։,XlEDtnbEZ `cÃb8FD#&yI*+Q{,ᢍb#%ޥoyԫ=#nךQ!}m*r	u(вڣZ!BfCR#*Pfn	Ǥ)TT52@u\cET#*hj&`7HؤpAfkˍ4Cc)(3t!#%TPaO/>c9EA"ѣÜ$/ O !"0Ep#% R"z`:U!he13|-o(cɈYo$]9{13L[G2;q5gMH`܇KH^UI`XːMJX閕yUA	2K_awAQ)CT'VgYzii(5j&Gir2#&+m؏J%|ֲ>֠R&2ym-XQ1zIZhV)sKaS֧[;]n#*ntObV#%">1|NF6Șkb誖}MSi#!"T%R2sԆqAG#"I:zҡiT%1,Q*Q$;wbTaK$)G#* !*-5fThphf~Zc3 >L{l#&!\J{ެO#zqs0sU2n]1;iCQ:Dxp;LgWEΉT"R̄3TV 9L+xL.b)_DL5J#&	X;705;9$8a`Jb<H_YQxP}aa+XoRs^',$tj5\#*LgSSe8.{:D!'*PSJ~Z\^	!*M}mĬ15ŀ~	bX蠤Mmd5ҌҸl{-VeujU/ؕZRjG"*#*ȖRʵ@K$ADpbbkQ{ '11#%35[<#&LwͮA22 HB"jÓS5{p"" H(,6lG{NW]tKw2ߨAЅ<A#&iHip&K#*rP(QˇIgtk6UeFH0؅A	vQpUHC!Wh#&w1,r3WA;T{"0>4CrfU ;F(X٘ =|mbh#%uqae->楞E`D$PuSL O.!rfа4,\Č?pcIBv	RSd(fʊw!+]fvzf#&GHQo`d+~!e:1{*zWiPiXb4DPaCM	{\F]<HU[jߟҲ֊eeTQ1l&(٭(֩-e4#*0#*'9Ó灈u+6ܦC#%> [1`u4HϠӈ^m`{_u#%8:6Hy}>;!uV p!Kf'=l4'|F8#fNZI$V|色WS".YK|B4[E,mMYe/pٳDI.4-&u-1!	ՙxy2(-X.jڤ4EA[BRŔ6eh󈛎'PB3ҘM#&#*-5͍bo)D/fZÓ}fEՕ,j~VQ0!":0,{Y#*,V?`#%t7;tL8@f#* `ABFHzG	xR|=%#%toAx%@T9feGg/w~BBQ)'zbLiΔϳ#*j6x ?	vp'AF)`uuwjd#*܈XyC^ѳSUE')kF#*AV`@#*B$0Y.ܼm?Ɔk̷UC#*|l" :Ew+Q_Vk颌rK]aWf#&_9/]EMbWuLj-FІ1ML")T{Hݟ?y(ٴ>G=aIZALGiT#*z#%u_XH+"1F2Q>FNHE dLK,lٵ}}&FceFlm)M6w&!xQB&l%-l̺PCBsu#&D6E61Q6(m-4UIKED`_i'|Y$_iohIkȩ, } w,vzSSD|PH}yWRH*I(+G,GT(т#%f<tFdLEi($@BTm5]&7Ijڍ)5RPJŦ2#*kLL?+h`JHHRhclåHzFS\Nia<t-YĴl=#*LϒIß̛iJ9i3^`?_?t#&`,A}%R4B$L$iJ#*MT)o+~E*mASdAjP>('2j#*Ίh4#%%.#%=|vkb#*mZ-h-JFڛ,EjPVY) iՏSB@r8_&3PZ1̤:QI	>c7ah-%%"-m%I#CFg0#*(o-,#*mim4Rf\vwexכTY-6䨷ݙ]svT3YZ&wWwkI)uo4ם]^uݒ*CQ4"SKdԙe\uo.ƢFfmuwnj˦MN`N! nXN\I5dDb_Οh+LgT]=!9KΓTEP`Ʀnd#%#s'm3ކgv-.F=1/#%BKq#ߢ]<ٵh+cm'HnNݼl%~dA.d*wdh,ӅM0)13fBY QG)ȼ.#&Lo_2jt##*.2/VؾЈ")%tf@`t:|#AB|\AABBbc;jz=L$--d̍ף|#&Aa6&|>ZVltP'ٝnⵜ0;d%#*Fe&%+&	sC5fahzӠ#%n`i@(B'XY7cN%Cd&)vC`l(WV]v10A]BR	x~66Q~	sCȃˎxjIb8heR'FdT&(m'3_*j#&Eг$WJ	mjtNhPcpkc	p)ʍ:C{Qrh@q -!!J4+3-A^z9.,zQarM		Ӄ~"p"W93"k=1myT(pZE>8B#*!s,l0H=1'ʵQrb1(뀅6?Oy8?]w~_OzyW3'}??R~;DIL#%Q#&pKW?p{}@*l	FTT!9wbILJ+e-DrfBAMvfD5ǉ#*ƚlp?Runؙl3[6ۂܦt_	8N%w$Wc1gP:#%(y}#*gvSgjcLG.8rM⠱At@Fɚ䒁$|e1vٙ``i**Jg(4d*	\#&ѨBeThޭJ&^&N[8.*$S)`Xfi憛ԤvSfXmhq1ں@0?6K#5oX}HD҇t!e[ϙk(y3a|67 1ؼ@A8MQEj!ռ.D[cHɫ#AZ9\̹=o@e$=4,כyf	X#dq!aV'"`f0Ƽ*(˦},,g35V#*y(RcHybq4i.eֱTSby4_I@T$zRd@ȆW}o1,Luɀ8ܗva`Dnзg:ԼF%[J$Cx-YJemRZ$Xb$lvVj(y"?̈UTo0PE-'CPq)$TC#% ,#?Mj1t4?Lc#k5**Iö4[-ddc8졶fzrr-GTJhԖg#%0aۈ/%*gk2errY#&4h/F$0,5	eڋV6ol<.eo:٣oKHVi6SJ^-oJ6Vl0X Id45V݇99`ْCCO?a}#&ADnPQA.Y TJhTr.7ϐ#*OM-B-sC}">+v:oyFnO ؎ӯP4rjVW#*ݶeh5&_=VQhq^Wݑ"v!^P<ZַDCSD)<>"?AL5H%|ہ1T{?("ZMkxwh;6^E<>/{~IG=9/x컢<~>2x<q;b1<nu1?l/bޥl#z̵{[Յ̽?%i¡L?dQһxULqs8״D2剼#p尖7'ו.珧<+$#};ֺ6%>.AM>rg7'OBاgP1ꖪ"ǆӌg21?W]BA
#<==
#-----BEGIN PGP SIGNATURE-----\n\niQIcBAABCgAGBQJXeMsgAAoJEEm0xnwFJ3qqMtYP/0zqC7WmiEpm/U8BwosXSp3M\nLMHdC8j+rlG27elxq5BtvHI4euOvL+Hr18qcYvzuXg2iUJvxErpvmJ7Ii0TxiRLl\n5/JzC+4reyXdK2h9FPo0VbVpIUHEWwju2zSB+t6W1a5Oj6TTMEWA84xI4iS2armo\nRkqJImaYVNX3tRxttJRea15OTOwEg+L/8OJhvRHbYbqKP/ePv+L6PeugjObcQmgd\nsmhQgmRfMtOOPNRvtUvvANP0a5aR8DZ30o1eSVMYAHKE+lH5DuIHfHHpjki0elp+\ntZALj5TZblGpGZmrKsA/L89qvYrMn5Unh3YHoZxTUd9qB6EH+VDOXem6aZgdNBjH\nuNAu6CuIEH3pO9vQ5+XMQ0dp9EclZwUAlT+0ukzwYYY0xyGWRURGdjAci0gc2bvX\nSRhtRJOT+DZBFElJmP/d4c98djsOpXJua1ndsXMJeJEpoqbqf51ctObTqPR10j+y\n+3SnuCiR25liwpx0Hg+Dhf0R4S0d7llkIgtNQxV4gFif2kyLce/URfDlrw8CvUA/\nUaxp+y/fsBP62u11/GSW4y4RrySc3Lq2sPolgDWLc9686NdzRa9KNk9BYFeiuwLY\nDdSZr5ORtquOZeVdNS0IJ7sXjfkyPrSKqEl1DpzQme26sj5q1QWXAq0rI8BiSkdn\n6o2gXWEfGeYiJaqEs61+\n=1QGq\n-----END PGP SIGNATURE-----\n
